import SecureBlackbox.PGP.*;

import java.awt.BorderLayout;
import javax.swing.JLabel;
import javax.swing.SwingConstants;
import java.awt.Dimension;
import java.awt.Font;

import javax.swing.JFileChooser;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JTree;
import javax.swing.JButton;
import javax.swing.JSeparator;
import javax.swing.event.TreeSelectionEvent;
import javax.swing.event.TreeSelectionListener;
import javax.swing.filechooser.FileNameExtensionFilter;
import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.DefaultTreeModel;
import javax.swing.tree.TreePath;

import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;


public class MimeViewer_OptionsPGPMime extends MimeViewer_PlugControl {
	
	private static final long serialVersionUID = 1L;
	public static TElPGPKeyring Keyring = null;
	private DefaultMutableTreeNode rootNode;
	private JTree tvKeys;
	private JLabel lblKeyId;
	private JLabel lblKeyFp;
	private JLabel lblAlgorithm;
	
	public MimeViewer_OptionsPGPMime() {
		setLayout(new BorderLayout(0, 0));
		
		JLabel lblPgp = new JLabel("PGP");
		lblPgp.setFont(new Font("Tahoma", Font.BOLD, 14));
		lblPgp.setPreferredSize(new Dimension(19, 30));
		lblPgp.setHorizontalAlignment(SwingConstants.CENTER);
		add(lblPgp, BorderLayout.NORTH);
		
		JPanel panel = new JPanel();
		panel.setPreferredSize(new Dimension(10, 70));
		add(panel, BorderLayout.SOUTH);
		panel.setLayout(null);
		
		lblKeyId = new JLabel("Key ID:");
		lblKeyId.setBounds(10, 11, 703, 14);
		panel.add(lblKeyId);
		
		lblKeyFp = new JLabel("Key FP:");
		lblKeyFp.setBounds(10, 29, 703, 14);
		panel.add(lblKeyFp);
		
		lblAlgorithm = new JLabel("Algorithm:");
		lblAlgorithm.setBounds(10, 47, 703, 14);
		panel.add(lblAlgorithm);
		
		rootNode = new DefaultMutableTreeNode("#");
		tvKeys = new JTree();
		tvKeys.addTreeSelectionListener(new TreeSelectionListener() {
			public void valueChanged(TreeSelectionEvent e) {
				treeAfterSelect(e);
			}
		});
		
		tvKeys.setModel(new DefaultTreeModel(rootNode));
		tvKeys.setRootVisible(false);
		tvKeys.setShowsRootHandles(true);
		add(tvKeys, BorderLayout.CENTER);

		CustomTreeCellRenderer render = new CustomTreeCellRenderer();
		tvKeys.setCellRenderer(render);

		
		JPanel panel_1 = new JPanel();
		panel_1.setPreferredSize(new Dimension(140, 10));
		add(panel_1, BorderLayout.EAST);
		panel_1.setLayout(null);
		
		JButton btnAddKey = new JButton("Add Key");
		btnAddKey.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				addClick();
			}
		});
		btnAddKey.setBounds(10, 11, 120, 23);
		panel_1.add(btnAddKey);
		
		JButton btnRemoveKey = new JButton("Remove Key");
		btnRemoveKey.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				removeClick();
			}
		});
		btnRemoveKey.setBounds(10, 42, 120, 23);
		panel_1.add(btnRemoveKey);
		
		JButton btnSaveKey = new JButton("Save Key");
		btnSaveKey.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				saveClick();
			}
		});
		btnSaveKey.setBounds(10, 76, 120, 23);
		panel_1.add(btnSaveKey);
		
		JButton btnClearKeyring = new JButton("Clear Keyring");
		btnClearKeyring.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				clearClick();
			}
		});
		btnClearKeyring.setBounds(10, 110, 120, 23);
		panel_1.add(btnClearKeyring);
		
		JSeparator separator = new JSeparator();
		separator.setBounds(10, 104, 120, 2);
		panel_1.add(separator);
		init();
	}

	protected void treeAfterSelect(TreeSelectionEvent e) {
	       TreePath p = e.getPath();

	       Object node = p.getLastPathComponent();
	       
	       if (node instanceof DefaultMutableTreeNode)
	       {
	    	   Object tag = ((NodeData)((DefaultMutableTreeNode)node).getUserObject()).getTag();
	    	   if ((tag instanceof TElPGPPublicKey) || (tag instanceof TElPGPPublicSubkey))
	    	   {
					TElPGPCustomPublicKey PubKey = (TElPGPCustomPublicKey)tag;
					lblKeyId.setText("Key ID: " + SBPGPUtils.keyID2Str(PubKey.keyID(), false));
					lblKeyFp.setText("Key FP: " + SBPGPUtils.keyFP2Str(PubKey.keyFP()));
					lblAlgorithm.setText("Algorithm: " + SBPGPUtils.pkAlg2Str(PubKey.getPublicKeyAlgorithm()) +
						" (" + PubKey.getBitsInKey() + " bits)");
					return;

	    	   } else if (tag instanceof TElPGPSignature) {
	    		   lblKeyId.setText("Signing Key ID: " + SBPGPUtils.keyID2Str(((TElPGPSignature)tag).signerKeyID(), false));
	    		   lblKeyFp.setText("");
	    		   lblAlgorithm.setText("");
					return;
	    	   }
	       }
	}

	protected void clearClick() {
		Keyring.clear();
		RepaintKeyring();
	}

	protected void saveClick() {
		
		Object Tag = getSelectedNodeTag();
		if (Tag == null) return;
		
		
		if (Tag instanceof TElPGPPublicKey)
		{
			TElPGPPublicKey key = (TElPGPPublicKey) Tag;
			String fileName = getSaveFileName();

			if (fileName.length() == 0) 
				return;
		
			if (key.getSecretKey() != null) 
				key.getSecretKey().saveToFile(fileName , false);
			else
				key.saveToFile(fileName, false);

			showMessage("Key successfully saved", "MIME Viewer");
		}
	}

	static void showMessage(String caption, String msg) {
		JOptionPane.showMessageDialog(null, msg, caption, JOptionPane.INFORMATION_MESSAGE);
	}	
	
	protected void removeClick() {
		if (showConfirmMessage("Are you sure you wish to remove this key from keyring?", "MIME Viewer"))
			return;

		Object Tag = getSelectedNodeTag();
		if (Tag == null) return;
		if (Tag instanceof TElPGPPublicKey)
		{
			TElPGPPublicKey key = (TElPGPPublicKey) Tag;
			if (key.getSecretKey() != null) 
				Keyring.removeSecretKey(key.getSecretKey());
			else
				Keyring.removePublicKey(key, false);
		
			RepaintKeyring();
		}
	}
	
	Object getSelectedNodeTag() {
	    TreePath[] paths = tvKeys.getSelectionPaths();

	    if (paths.length > 0)
	    {
	    	NodeData node = getFirstNode(paths);
	    	if (node == null) return null;
	    	return node.getTag();
	    }
	    
	    return null;	
	}

	NodeData getFirstNode(TreePath[] paths){
		Object obj = ((DefaultMutableTreeNode)paths[0].getLastPathComponent()).getUserObject();
		if (obj instanceof NodeData)
			return (NodeData)obj;
		return null;
	}
	
	private boolean showConfirmMessage(String msg, String cap) {
		return (JOptionPane.showConfirmDialog(this, msg, cap, JOptionPane.YES_NO_OPTION) == JOptionPane.YES_OPTION);
	}

	protected void addClick() {
		TElPGPKeyring TempKeyring;
		String fileName = getOpenFileName();
		if (fileName.length() > 0)
		{
			TempKeyring = new TElPGPKeyring();
			try
			{
				TempKeyring.load(fileName, "", true);
				TempKeyring.exportTo(Keyring);
			}
			finally
			{
				TempKeyring.Free();
			}
			RepaintKeyring();
		}
	}

	private String getSaveFileName() {
		JFileChooser fc = new JFileChooser();
	
	    int returnVal = fc.showSaveDialog(this);
		if (returnVal == JFileChooser.APPROVE_OPTION)
	        return fc.getSelectedFile().getPath();

	    return "";
	}
	
	private String getOpenFileName() {
		JFileChooser fc = new JFileChooser();

		FileNameExtensionFilter filter = new FileNameExtensionFilter("OpenPGP keyring files", "pkr", "skr", "pgp", "gpg", "asc");
	    fc.setFileFilter(filter);		
	
	    int returnVal = fc.showOpenDialog(this);
		if (returnVal == JFileChooser.APPROVE_OPTION)
	        return fc.getSelectedFile().getPath();

	    return "";
	}

	private void init() {
		if (Keyring == null)
		{
			Keyring = new TElPGPKeyring();
			try
			{
				Keyring.load("res\\EldoS MIMEBlackbox Demo.asc", "", true);
			}
			catch(Exception ex)
			{
				// do nothing
			}
			RepaintKeyring();
		}		
	}

	private static void expandTree(JTree tree) {	
		tree.setRootVisible(true);

		int row = 0;
	    while (row < tree.getRowCount()) {
	      tree.expandRow(row);
	      row++;
	    }
	    
		tree.setRootVisible(false);
	}
	
	private void RepaintKeyring() {
		try
		{
			clearTree();

			int i, j, k;
			String S;
			DefaultMutableTreeNode Node, SubNode, SigNode;

			for (i = 0; i < Keyring.getPublicCount(); i++)
			{
				if (Keyring.getPublicKeys(i).getUserIDCount() > 0)
					S = Keyring.getPublicKeys(i).getUserID(0).getName();
				else
					S = "<unnamed key>";

				Node = new DefaultMutableTreeNode(S);
				Node.setUserObject(new NodeData(Keyring.getPublicKeys(i), S));
				rootNode.add(Node);
				
				for (j = 0; j < Keyring.getPublicKeys(i).getUserIDCount(); j++)
				{
					SubNode = new DefaultMutableTreeNode(Keyring.getPublicKeys(i).getUserID(j).getName());
					SubNode.setUserObject(new NodeData(Keyring.getPublicKeys(i).getUserID(j), Keyring.getPublicKeys(i).getUserID(j).getName()));
					Node.add(SubNode);
					for (k = 0; k < Keyring.getPublicKeys(i).getUserID(j).getSignatureCount(); k++)
					{
						SigNode = new DefaultMutableTreeNode("signature");
						SigNode.setUserObject(new NodeData(Keyring.getPublicKeys(i).getUserID(j).getSignature(k),"signature"));
						SubNode.add(SigNode);
					}
				}

				for (j = 0; j < Keyring.getPublicKeys(i).getSubkeyCount(); j++)
				{
					SubNode = new DefaultMutableTreeNode("Subkey");
					SubNode.setUserObject(new NodeData(Keyring.getPublicKeys(i).getSubkey(j), "Subkey"));
					Node.add(SubNode);
					
					for (k = 0; k < Keyring.getPublicKeys(i).getSubkey(j).getSignatureCount(); k++)
					{
						SigNode = new DefaultMutableTreeNode("signature");
						SigNode.setUserObject(new NodeData(Keyring.getPublicKeys(i).getSubkey(j).getSignatures(k), "signature"));
						SubNode.add(SigNode);
					}
				}
			}
		}
		finally
		{
			expandTree(tvKeys);
		}
	}

	private void clearTree() {
		rootNode.removeAllChildren();
	}
	
	public String GetCaption()
	{
		return "PGP/MIME Options";
	}
}
