import java.util.Vector;

import javax.swing.table.AbstractTableModel;

import SecureBlackbox.Office.TElOfficeBinarySignedEntry;
import SecureBlackbox.Office.TElOfficeBinaryXMLSignatureHandler;
import SecureBlackbox.Office.TElOfficeCustomSignatureHandler;
import SecureBlackbox.Office.TElOpenOfficeSignatureHandler;
import SecureBlackbox.Office.TSBOfficeBinaryEntryValidationStatus;
import SecureBlackbox.Office.TSBOpenOfficeEntryValidationStatus;

public class EntriesTableModel extends AbstractTableModel {
	private static final long serialVersionUID = 2771284071470299948L;
	public static final int PATH_INDEX = 0;
	public static final int STATE_INDEX = 1;
	
	protected Vector<EntriesDataRow> dataVector = new Vector<EntriesDataRow>();
	
    protected String[] columnNames = {
    		"Path", 
    		"State", 
    };
    
	public String getColumnName(int column) {
        return columnNames[column];
    }
    
	public int getColumnCount() {
		return columnNames.length;
	}

	public int getRowCount() {
		return dataVector.size();
	}

    public Object getValueAt(int row, int col) {
        EntriesDataRow record = (EntriesDataRow)dataVector.get(row);
        switch (col) {
        	case PATH_INDEX:
	        	 return record.Path;
            case STATE_INDEX:
               return record.State;
            default:
               return new Object();
        }
    }

	public void clear() {
		dataVector.removeAllElements();
	}

    public void addRow(String path, String state, Object tag){
    	dataVector.addElement(new EntriesDataRow(path, state, tag));
    	fireTableDataChanged();
    }
    
	public void fill(TElOfficeCustomSignatureHandler _Handler) {
        if (_Handler instanceof TElOfficeBinaryXMLSignatureHandler)
        {
            TElOfficeBinaryXMLSignatureHandler BinaryHandler = (TElOfficeBinaryXMLSignatureHandler)_Handler;
            for (int i = 0; i < BinaryHandler.getSignedEntryCount(); i++)
            {
                String path = BinaryHandler.getSignedEntry(i).getPath();
                TElOfficeBinarySignedEntry tag = BinaryHandler.getSignedEntry(i);
                TSBOfficeBinaryEntryValidationStatus status = BinaryHandler.getSignedEntry(i).getValidationStatus();

                String state = "Unknown";
                if (status == TSBOfficeBinaryEntryValidationStatus.bevsValid) state = "Valid";
                if (status == TSBOfficeBinaryEntryValidationStatus.bevsInvalidDigest) state = "Invalid Digest";
                if (status == TSBOfficeBinaryEntryValidationStatus.bevsInvalidSignature) state = "Invalid Signature";
                if (status == TSBOfficeBinaryEntryValidationStatus.bevsEntryIsMissing) state = "Entry Is Missing";
                if (status == TSBOfficeBinaryEntryValidationStatus.bevsEntryNotSigned) state = "Entry Not Signed";
                
                addRow(path, state, tag);
            }
        }
        else if (_Handler instanceof TElOpenOfficeSignatureHandler)
        {
            TElOpenOfficeSignatureHandler ODFHandler = (TElOpenOfficeSignatureHandler)_Handler;
            for (int i = 0; i < ODFHandler.getSignedEntryCount(); i++)
            {
                String path = ODFHandler.getSignedEntry(i).getPath();
                Object tag = ODFHandler.getSignedEntry(i);
                
                TSBOpenOfficeEntryValidationStatus status = ODFHandler.getSignedEntry(i).getValidationStatus();
                String state = "Unknown";
            	if (status == TSBOpenOfficeEntryValidationStatus.oevsValid) state = "Valid";
            	if (status == TSBOpenOfficeEntryValidationStatus.oevsInvalidDigest) state = "Invalid Digest";
            	if (status == TSBOpenOfficeEntryValidationStatus.oevsInvalidSignature) state = "Invalid Signature";
            	if (status == TSBOpenOfficeEntryValidationStatus.oevsEntryIsMissing) state = "Entry Is Missing";
            	
            	addRow(path, state, tag);
            }
        }		
	}
}