import java.awt.BorderLayout;
import java.awt.FlowLayout;
import java.util.Date;

import javax.swing.JButton;
import javax.swing.JDialog;
import javax.swing.JPanel;
import javax.swing.border.EmptyBorder;
import javax.swing.JLabel;
import javax.swing.JComboBox;
import javax.swing.JOptionPane;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;

import SecureBlackbox.Base.SBUtils;
import SecureBlackbox.Base.TElCustomCertStorage;
import SecureBlackbox.Base.TElX509Certificate;
import SecureBlackbox.Base.TElX509CertificateValidator;
import SecureBlackbox.Base.TElX509CertificateValidatorResult;
import SecureBlackbox.Base.TSBCertificateValidity;
import SecureBlackbox.Office.*;

import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import java.awt.event.ItemListener;
import java.awt.event.ItemEvent;


public class SignaturesInfoForm extends JDialog {

	private static final long serialVersionUID = 1L;
	private final JPanel contentPanel = new JPanel();
	private JComboBox<String> cmbSignatures;
	private JTextArea tbInfo;

    private TElOfficeDocument _OfficeDocument = null;
    private TElX509Certificate _SignerCertificate;
    private TElCustomCertStorage _AdditionalCertificates;
    private Date _SigningTime;
    private TElX509CertificateValidator CertificateValidator;
    private JButton btnDetails;
    private JButton btnValidateCertificate;
	
	/**
	 * Launch the application.
	 */
	public static void main(String[] args) {
		try {
			SignaturesInfoForm dialog = new SignaturesInfoForm();
			dialog.setDefaultCloseOperation(JDialog.DISPOSE_ON_CLOSE);
			dialog.setVisible(true);
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	/**
	 * Create the dialog.
	 */
	public SignaturesInfoForm() {
		setTitle("Signatures Info");
		setBounds(100, 100, 486, 300);
		getContentPane().setLayout(new BorderLayout());
		contentPanel.setBorder(new EmptyBorder(5, 5, 5, 5));
		getContentPane().add(contentPanel, BorderLayout.CENTER);
		contentPanel.setLayout(null);
		{
			JLabel lblSignatures = new JLabel("Signatures");
			lblSignatures.setBounds(10, 11, 74, 14);
			contentPanel.add(lblSignatures);
		}
		{
			cmbSignatures = new JComboBox<String>();
			cmbSignatures.addItemListener(new ItemListener() {
				public void itemStateChanged(ItemEvent e) {
					cmbSignatures_SelectedIndexChanged();
				}
			});
			cmbSignatures.setBounds(94, 7, 146, 22);
			contentPanel.add(cmbSignatures);
		}
		{
			JButton btnValidate = new JButton("Validate");
			btnValidate.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					validateClick();
				}
			});
			btnValidate.setBounds(250, 7, 74, 23);
			contentPanel.add(btnValidate);
		}
		{
			btnValidateCertificate = new JButton("Validate Certificate");
			btnValidateCertificate.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					validateCertificateClick();
				}
			});
			btnValidateCertificate.setBounds(334, 7, 134, 23);
			contentPanel.add(btnValidateCertificate);
		}
		
		JScrollPane scrollPane = new JScrollPane();
		scrollPane.setBounds(10, 36, 458, 193);
		contentPanel.add(scrollPane);
		
		tbInfo = new JTextArea();
		scrollPane.setViewportView(tbInfo);
		{
			JPanel buttonPane = new JPanel();
			buttonPane.setLayout(new FlowLayout(FlowLayout.RIGHT));
			getContentPane().add(buttonPane, BorderLayout.SOUTH);
			{
				btnDetails = new JButton("Details...");
				btnDetails.addActionListener(new ActionListener() {
					public void actionPerformed(ActionEvent e) {
						detailsClick();
					}
				});
				btnDetails.setActionCommand("Cancel");
				buttonPane.add(btnDetails);
			}
			{
				JButton okButton = new JButton("OK");
				okButton.setActionCommand("OK");
				okButton.addActionListener(new ActionListener() {
					public void actionPerformed(ActionEvent e) {
						dispose();
					}
				});
				buttonPane.add(okButton);
				getRootPane().setDefaultButton(okButton);
			}
		}
	}
	
	protected void validateCertificateClick() {
        if (_SignerCertificate == null)
        {
            showErrorMessage("No certificate");
            return;
        }

		TElX509CertificateValidatorResult result = new TElX509CertificateValidatorResult();
		// For proper CRL and OCSP validation please read instructions in 
		// description of ElX509CertificateValidator class in the help file
        CertificateValidator.validate(_SignerCertificate, _AdditionalCertificates, false, _SigningTime, result );
        if (result.Validity == TSBCertificateValidity.cvOk)
            showMessage("Signer certificate is valid.");
        else if (result.Validity == TSBCertificateValidity.cvSelfSigned)
            showMessage("The signer signature is signed by self-signed certificate which was not previously trusted");
        else
        {
            if (result.Validity == TSBCertificateValidity.cvChainUnvalidated)
                showErrorMessage("Signer certificate chain could not be validated completely.");
            else
                showErrorMessage("Signer certificate is not valid.");
        }
	}

	protected void detailsClick() {
        if ((cmbSignatures.getSelectedIndex() < 0) || (cmbSignatures.getSelectedIndex() >= _OfficeDocument.getSignatureHandlerCount()))
            return;

        TElOfficeCustomSignatureHandler Handler = _OfficeDocument.getSignatureHandler(cmbSignatures.getSelectedIndex());
        if (Handler.getClass().getName().indexOf("TElOfficeOpenXMLBaseSignatureHandler") >= 0)
        {
            SignedPartsInfoForm frmSignedPartsInfo = new SignedPartsInfoForm();
            {
                frmSignedPartsInfo.SetSignatureHandler(Handler);
                frmSignedPartsInfo.setVisible(true);
            }
        }
        else if ((Handler.getClass().getName().indexOf("TElOfficeBinaryXMLSignatureHandler") >= 0) || (Handler.getClass().getName().indexOf("TElOpenOfficeSignatureHandler") >= 0))
        {
            SignedEntriesInfoForm frmSignedEntriesInfo = new SignedEntriesInfoForm();
            {
                frmSignedEntriesInfo.SetSignatureHandler(Handler);
                frmSignedEntriesInfo.setVisible(true);
            }
        }
    }

	protected void cmbSignatures_SelectedIndexChanged() {
        btnDetails.setEnabled(false);
        btnValidateCertificate.setEnabled(false);
        if ((cmbSignatures.getSelectedIndex() < 0) || (cmbSignatures.getSelectedIndex() >= _OfficeDocument.getSignatureHandlerCount()))
            return;

        TElOfficeCustomSignatureHandler Handler = _OfficeDocument.getSignatureHandler(cmbSignatures.getSelectedIndex());
        String s = "";
        s = s + "The document is signed using \"" + Handler.getClass().getName() + "\" handler" + "\r\n" + "\r\n";
        if (Handler.getClass().getName().indexOf("TElOfficeOpenXMLInvalidSignatureHandler") >= 0)
        {
            s = s + "Error message: " + ((TElOfficeOpenXMLInvalidSignatureHandler)Handler).getErrorMessage();
        }
        else if (Handler.getClass().getName().indexOf("TElOfficeBinaryInvalidSignatureHandler") >= 0)
        {
            s = s + "Error message: " + ((TElOfficeBinaryInvalidSignatureHandler)Handler).getErrorMessage();
        }
        else if (Handler.getClass().getName().indexOf("TElOfficeOpenXMLBaseSignatureHandler") >= 0)
        {
            s = s + "Signature time: " + ((TElOfficeOpenXMLBaseSignatureHandler)Handler).getSignatureTime().getValueUTC().toString() + "\r\n";
            if (Handler.getClass().getName().indexOf("TElOfficeOpenXMLSignatureHandler") >= 0)
            {
                TElOfficeOpenXMLSignatureHandler OpenXMLHandler = (TElOfficeOpenXMLSignatureHandler)Handler;
                if (OpenXMLHandler.getSignatureLine() != null)
                {
                    s = s + "Signature is attached to the signature line:" + "\r\n";
                    s = s + " - ID: " + OpenXMLHandler.getSignatureLine().getID() + "\r\n";
                    if (OpenXMLHandler.getSignatureLine().getSuggestedSigner().length() > 0)
                        s = s + " - Suggested Signer: " + OpenXMLHandler.getSignatureLine().getSuggestedSigner() + "\r\n";

                    if (OpenXMLHandler.getSignatureLine().getSuggestedSigner2().length() > 0)
                        s = s + " - Suggested Signer2: " + OpenXMLHandler.getSignatureLine().getSuggestedSigner2() + "\r\n";

                    if (OpenXMLHandler.getSignatureLine().getSuggestedSignerEmail().length() > 0)
                        s = s + " - Suggested Signer Email: " + OpenXMLHandler.getSignatureLine().getSuggestedSignerEmail() + "\r\n";

                    if (OpenXMLHandler.getSignatureLine().getSigningInstructions().length() > 0)
                        s = s + " - Signing Instructions: " + OpenXMLHandler.getSignatureLine().getSigningInstructions() + "\r\n";

                    if (OpenXMLHandler.getSignatureLine().getAdditionalSignatureInfo().length() > 0)
                        s = s + " - Additional Signature Info: " + OpenXMLHandler.getSignatureLine().getAdditionalSignatureInfo() + "\r\n";
                    // ...
                }

                s = s + FormatSignatureInfoV1(OpenXMLHandler.getSignatureInfoV1()) + "\r\n";
                if (OpenXMLHandler.isDocumentSigned())
                    s = s + "Document content is signed" + "\r\n";
                else
                    s = s + "Document content is partially signed" + "\r\n";
            }
            else if (Handler.getClass().getName().indexOf("TElOfficeOpenXPSSignatureHandler") >= 0)
            {
                TElOfficeOpenXPSSignatureHandler OpenXPSHandler = (TElOfficeOpenXPSSignatureHandler)Handler;
                if (OpenXPSHandler.getSignatureDefinition() != null)
                {
                    s = s + "Signature is attached to the signature definition:" + "\r\n";
                    s = s + " - Spot ID: " + OpenXPSHandler.getSignatureDefinition().getSpotID() + "\r\n";
                    if (OpenXPSHandler.getSignatureDefinition().getIntent().length() > 0)
                        s = s + " - Intent: " + OpenXPSHandler.getSignatureDefinition().getIntent() + "\r\n";

                    if (OpenXPSHandler.getSignatureDefinition().getSignerName().length() > 0)
                        s = s + " - Signer Name: " + OpenXPSHandler.getSignatureDefinition().getSignerName() + "\r\n";

                    if (OpenXPSHandler.getSignatureDefinition().getSigningLocation().length() > 0)
                        s = s + " - Signing Location: " + OpenXPSHandler.getSignatureDefinition().getSigningLocation() + "\r\n";

                    if (OpenXPSHandler.getSignatureDefinition().getSignBy().toString().length() > 0)
                        s = s + " - Sign By: " + OpenXPSHandler.getSignatureDefinition().getSignBy().toString() + "\r\n";

                    if (OpenXPSHandler.getSignatureDefinition().getPageURI().length() > 0)
                    {
                        s = s + " - Page URI: " + OpenXPSHandler.getSignatureDefinition().getPageURI() + "\r\n";
                    }
                }

                if (!OpenXPSHandler.isDocumentSigned())
                    s = s + "\r\n" + "This is NOT an OpenXPS signature" + "\r\n";

                if (OpenXPSHandler.isSignatureOriginSigned())
                    s = s + "Signature origin is signed" + "\r\n";
            }

            if (((TElOfficeOpenXMLBaseSignatureHandler)Handler).isCorePropertiesSigned())
                s = s + "Document properties are signed" + "\r\n";

            btnDetails.setEnabled(true);
        }
        else if (Handler.getClass().getName().indexOf("TElOfficeBinaryCryptoAPISignatureHandler") >= 0)
        {
            s = s + "Sign time: " + ((TElOfficeBinaryCryptoAPISignatureHandler)Handler).getSignTime().toString() + "\r\n";
            if (((TElOfficeBinaryCryptoAPISignatureHandler)Handler).getExpireTime()	!= SBUtils.dateTimeFromOADate(0))
                s = s + "Expire time: " + ((TElOfficeBinaryCryptoAPISignatureHandler)Handler).getExpireTime().toString() + "\r\n";
        }
        else if (Handler.getClass().getName().indexOf("TElOfficeBinaryXMLSignatureHandler") >= 0)
        {
            TElOfficeBinaryXMLSignatureHandler BinXMLHandler = (TElOfficeBinaryXMLSignatureHandler)Handler;
            s = s + "Signature time: " + BinXMLHandler.getSignatureTime().getValueUTC().toString() + "\r\n";
            s = s + FormatSignatureInfoV1(BinXMLHandler.getSignatureInfoV1()) + "\r\n";
            btnDetails.setEnabled(true);
        }
        else if (Handler.getClass().getName().indexOf("TElOpenOfficeSignatureHandler") >= 0)
        {
            TElOpenOfficeSignatureHandler ODFHandler = (TElOpenOfficeSignatureHandler)Handler;
            s = s + "Signature time: " + ODFHandler.getSignatureTime().getValueUTC().toString() + "\r\n";
            if (ODFHandler.isDocumentSigned())
                s = s + "Document content is signed" + "\r\n";
            else
                s = s + "Document content is partially signed" + "\r\n";

            btnDetails.setEnabled(true);
        }

        tbInfo.setText(s);
	}

    private String FormatSignatureInfoV1(TElOfficeXMLSignatureInfoV1 SignatureInfoV1)
    {
        String Result = "";
        if (SignatureInfoV1.getIncluded())
        {
            Result = Result + "Signature info:\r\n";
            if (SignatureInfoV1.getDelegateSuggestedSigner().length() > 0)
                Result = Result + " - Delegate Suggested Signer: " + SignatureInfoV1.getDelegateSuggestedSigner() + "\r\n";

            if (SignatureInfoV1.getDelegateSuggestedSigner2().length() > 0)
                Result = Result + " - Delegate Suggested Signer2: " + SignatureInfoV1.getDelegateSuggestedSigner2() + "\r\n";

            if (SignatureInfoV1.getDelegateSuggestedSignerEmail().length() > 0)
                Result = Result + " - Delegate Suggested Signer Email: " + SignatureInfoV1.getDelegateSuggestedSignerEmail() + "\r\n";

            if (SignatureInfoV1.getSignatureComments().length() > 0)
                Result = Result + " - Signature Comments: " + SignatureInfoV1.getSignatureComments() + "\r\n";

            if (SignatureInfoV1.getSignatureText().length() > 0)
                Result = Result + " - Signature Text: " + SignatureInfoV1.getSignatureText() + "\r\n";

            if (SignatureInfoV1.getOfficeVersion().length() > 0)
                Result = Result + " - Office Version: " + SignatureInfoV1.getOfficeVersion() + "\r\n";

            if (SignatureInfoV1.getWindowsVersion().length() > 0)
                Result = Result + " - Windows Version: " + SignatureInfoV1.getWindowsVersion() + "\r\n";

            Result = Result + String.format(" - Monitor(s): %d, Resolution: %d x %d px, %d bits" + "\r\n",  SignatureInfoV1.getMonitors(), SignatureInfoV1.getHorizontalResolution(), SignatureInfoV1.getVerticalResolution(), SignatureInfoV1.getColorDepth());
            // ...
        }

        return Result;
    }	
	void showMessage(String msg) {
		JOptionPane.showMessageDialog(null, msg, getTitle(), JOptionPane.INFORMATION_MESSAGE);
	}	

	protected void validateClick() {
        if ((cmbSignatures.getSelectedIndex() < 0) || (cmbSignatures.getSelectedIndex() >= _OfficeDocument.getSignatureHandlerCount()))
        {
            showMessage("Please select a signature to validate");
            return;
        }

        TElOfficeCustomSignatureHandler Handler = _OfficeDocument.getSignatureHandler(cmbSignatures.getSelectedIndex());
        if (Handler instanceof TElOfficeOpenXMLBaseSignatureHandler)
        {
            TSBOfficeOpenXMLSignatureValidationStatus ValidationStatus = ((TElOfficeOpenXMLBaseSignatureHandler)Handler).validate();
            if (ValidationStatus == TSBOfficeOpenXMLSignatureValidationStatus.svsValid)
                showMessage("Signature is VALID");
            else if (ValidationStatus == TSBOfficeOpenXMLSignatureValidationStatus.svsValidButNotParts)
                showErrorMessage("Signature integrity is ok, but some of parts are INVALID");
            else
            	showErrorMessage("Signature is INVALID");

            _SignerCertificate = ((TElOfficeOpenXMLBaseSignatureHandler)Handler).getSignerCertificate();
            _AdditionalCertificates = ((TElOfficeOpenXMLBaseSignatureHandler)Handler).getCertificates();
            _SigningTime = ((TElOfficeOpenXMLBaseSignatureHandler)Handler).getSignatureTime().getValueUTC();
        }
        else if (Handler instanceof TElOfficeBinaryCryptoAPISignatureHandler)
        {
            TSBOfficeBinarySignatureValidationStatus ValidationStatus = ((TElOfficeBinaryCryptoAPISignatureHandler)Handler).validate();
            if (ValidationStatus == TSBOfficeBinarySignatureValidationStatus.bsvsValid)
            	showMessage("Signature is VALID");
            else
            	showErrorMessage("Signature is INVALID");

            _SignerCertificate = ((TElOfficeBinaryCryptoAPISignatureHandler)Handler).getCertificate();
            _AdditionalCertificates = ((TElOfficeBinaryCryptoAPISignatureHandler)Handler).getIntermediateCertificatesStorage();
            _SigningTime = ((TElOfficeBinaryCryptoAPISignatureHandler)Handler).getSignTime();
        }
        else if (Handler instanceof TElOfficeBinaryXMLSignatureHandler)
        {
            TSBOfficeBinarySignatureValidationStatus ValidationStatus = ((TElOfficeBinaryXMLSignatureHandler)Handler).validate();
            if (ValidationStatus == TSBOfficeBinarySignatureValidationStatus.bsvsValid)
            	showMessage("Signature is VALID");
            else if (ValidationStatus == TSBOfficeBinarySignatureValidationStatus.bsvsValidButNotEntries)
            	showErrorMessage("Signature integrity is ok, but some of entries are INVALID");
            else
            	showErrorMessage("Signature is INVALID");

            _SignerCertificate = ((TElOfficeBinaryXMLSignatureHandler)Handler).getSignerCertificate();
            _AdditionalCertificates = ((TElOfficeBinaryXMLSignatureHandler)Handler).getCertificates();
            _SigningTime = ((TElOfficeBinaryXMLSignatureHandler)Handler).getSignatureTime().getValueUTC();
        }
        else if (Handler instanceof TElOpenOfficeSignatureHandler)
        {
            TSBOpenOfficeSignatureValidationStatus ODFValidationStatus = ((TElOpenOfficeSignatureHandler)Handler).validate();

            if (ODFValidationStatus == TSBOpenOfficeSignatureValidationStatus.osvsValid)
                showMessage("Signature is VALID");
            else if (ODFValidationStatus == TSBOpenOfficeSignatureValidationStatus.osvsValidButNotEntries)
            	showErrorMessage("Signature integrity is ok, but some of entries are INVALID");
            else
            	showErrorMessage("Signature is INVALID");

            _SignerCertificate = ((TElOpenOfficeSignatureHandler)Handler).getSignerCertificate();
            _AdditionalCertificates = ((TElOpenOfficeSignatureHandler)Handler).getCertificates();
            _SigningTime = ((TElOpenOfficeSignatureHandler)Handler).getSignatureTime().getValueUTC();
        }

        btnValidateCertificate.setEnabled(true);
	}

	void showErrorMessage(String msg){
		JOptionPane.showMessageDialog(null, msg, "Error", JOptionPane.ERROR_MESSAGE);
	}
	
	public void setDocument(TElOfficeDocument Document) {
        _OfficeDocument = Document;
        tbInfo.setText("Please select a signature");
        btnDetails.setEnabled(false);
        btnValidateCertificate.setEnabled(false);
        cmbSignatures.removeAllItems();
        for (int i = 0; i < _OfficeDocument.getSignatureHandlerCount(); i++)
        {
            cmbSignatures.addItem(String.format("Signature%d", (i + 1)));
        }
	}
}
