import java.util.Vector;
import javax.swing.table.*;
import SecureBlackbox.PGP.*;

public class KeyTableModel extends AbstractTableModel {
	static final long serialVersionUID = 6442378548025279799L;
	public static final int ST_INDEX   = 0;
	public static final int USER_INDEX = 1;
    public static final int KEY_INDEX  = 2;
    public static final int BITS_INDEX = 3;

    boolean publicKeys;
    protected String[] columnNames = {"", "User", "Key", "Bits" };
    protected Vector<KeyDataRow> dataVector = new Vector<KeyDataRow>();

    public KeyTableModel() {
    	
    }
    
 	public void init(TElPGPKeyring keys, boolean publicKeys) {
		if (publicKeys) LoadPub(keys);
		else LoadSec(keys);
 	} 
    
 	public String getColumnName(int column) {
        return columnNames[column];
    }

    private void LoadSec(TElPGPKeyring keys) {
 		for(int i = 0; i < keys.getSecretCount(); i++) 
 		{
 			String name = "<no name>";
 			TElPGPPublicKey key = keys.getSecretKeys(i).getPublicKey();
 			if (key.getUserIDCount() > 0) 
 				name = key.getUserID(0).getName();

 			String alg = SBPGPUtils.pkAlg2Str(key.getPublicKeyAlgorithm());
 			int bits = keys.getPublicKeys(i).getBitsInKey();
 			
 			Object tag = keys.getSecretKeys(i);
 			dataVector.add(new KeyDataRow(false, name, alg, bits, tag)); 			
 			setValueAt(name, i, USER_INDEX);
 			setValueAt(alg, i, KEY_INDEX);
 			setValueAt(bits, i, BITS_INDEX);
 		} 		
 	}

 	private void LoadPub(TElPGPKeyring keys) {
 		publicKeys = true;
 		for(int i = 0; i < keys.getPublicCount(); i++) 
 		{
 			String name = "<no name>";
 			if (keys.getPublicKeys(i).getUserIDCount() > 0) 
 				name = keys.getPublicKeys(i).getUserID(0).getName();

 			String alg = SBPGPUtils.pkAlg2Str(keys.getPublicKeys(i).getPublicKeyAlgorithm());
 			int bits = keys.getPublicKeys(i).getBitsInKey();
 			
 			Object tag = keys.getPublicKeys(i);
 			dataVector.add(new KeyDataRow(false, name, alg, bits, tag)); 			
 			setValueAt(name, i, USER_INDEX);
 			setValueAt(alg, i, KEY_INDEX);
 			setValueAt(bits, i, BITS_INDEX);
 		}
 	}     

 	public TElPGPKeyring getKeys() {
 		TElPGPKeyring result = new TElPGPKeyring();
 		for (int i = 0; i < dataVector.size(); i++){
 			KeyDataRow row = (KeyDataRow)dataVector.get(i);
 			if (row.Use){
 				if (publicKeys)
 					result.addPublicKey((TElPGPPublicKey) row.Tag);
 				else
 					result.addSecretKey((TElPGPSecretKey) row.Tag);
 			}
 		}
 		return result;
 	}
 	
 	public boolean isCellEditable(int row, int col) {
        return (col == ST_INDEX);
    }
 	
 	public Class<?> getColumnClass(int column) {
         switch (column) {
         	 case ST_INDEX:
         		 return Boolean.class;
             case KEY_INDEX:
             case USER_INDEX:
                return String.class;
             case BITS_INDEX:
            	 return Integer.class;
             default:
                return Object.class;
         }
     }

     public Object getValueAt(int row, int col) {
         KeyDataRow record = (KeyDataRow)dataVector.get(row);
         switch (col) {
	         case ST_INDEX:
	        	 return record.Use;
             case KEY_INDEX:
                return record.Key;
             case USER_INDEX:
                return record.Name;
             case BITS_INDEX:
                return record.Bits;
             default:
                return new Object();
         }
     }

     public void setValueAt(Object value, int row, int column) {
    	 KeyDataRow record = (KeyDataRow)dataVector.get(row);
         switch (column) {
	         case ST_INDEX:
	        	 record.Use = (Boolean)value;
	        	 break;
             case USER_INDEX:
                record.Name = (String)value;
                break;
             case KEY_INDEX:
                record.Key = (String)value;
                break;
             case BITS_INDEX:
                record.Bits = (Integer)value;
                break;
             default:
                System.out.println("invalid index");
         }
         fireTableCellUpdated(row, column);
     }

     public int getRowCount() {
         return dataVector.size();
     }

     public int getColumnCount() {
         return columnNames.length;
     }
 }