

import java.awt.EventQueue;

import javax.swing.JDialog;

import SecureBlackbox.Base.*;
import SecureBlackbox.HTTPClient.*;
import SecureBlackbox.PKI.*;
import SecureBlackbox.SSLCommon.*;

import javax.swing.JMenuBar;
import javax.swing.JMenu;
import javax.swing.JMenuItem;
import javax.swing.JSeparator;
import java.awt.BorderLayout;

import javax.swing.JFileChooser;
import javax.swing.JOptionPane;
import javax.swing.JPasswordField;
import javax.swing.JSplitPane;
import javax.swing.JScrollPane;
import javax.swing.JTextField;
import javax.swing.JTree;
import javax.swing.JTabbedPane;
import javax.swing.JPanel;
import javax.swing.JTable;
import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.DefaultTreeModel;
import javax.swing.tree.TreePath;

import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.io.File;
import java.io.FileInputStream;
import java.nio.charset.Charset;
import java.util.Date;

import javax.swing.JPopupMenu;

import org.freepascal.rtl.TObject;

import java.awt.Component;
import javax.swing.event.TreeSelectionListener;
import javax.swing.event.TreeSelectionEvent;


public class Main extends JDialog {

	private static final long serialVersionUID = 1L;
	private TElSignedCMSMessage FCMS;
	private TElFileStream FCMSStream = null;
	private TElFileStream FDataStream = null;
    private TElX509CertificateValidator FValidator = null;
	private String FLog;
	private JTable lvProps;
	private JTree tree;
	private DefaultMutableTreeNode rootNode;
	private JPopupMenu ctxMenu;
	
	/**
	 * Launch the application.
	 */
	public static void main(String[] args) {
		EventQueue.invokeLater(new Runnable() {
			public void run() {
				try {
					Main dialog = new Main();
					dialog.setDefaultCloseOperation(JDialog.DISPOSE_ON_CLOSE);
					dialog.setVisible(true);
				} catch (Exception e) {
					e.printStackTrace();
				}
			}
		});
	}

	/**
	 * Create the dialog.
	 */
	public Main() {
		setResizable(false);
		setTitle("EldoS CMS manager");
		setBounds(100, 100, 705, 607);
		getContentPane().setLayout(new BorderLayout(0, 0));
		
		JMenuBar menuBar = new JMenuBar();
		getContentPane().add(menuBar, BorderLayout.NORTH);
		
		JMenu mnFile = new JMenu("File");
		menuBar.add(mnFile);
		
		JMenuItem mntmNewCmsMessage = new JMenuItem("New CMS Message");
		mntmNewCmsMessage.addMouseListener(new MouseAdapter() {
			@Override
			public void mousePressed(MouseEvent e) {
				NewCMS();
			}
		});
		mnFile.add(mntmNewCmsMessage);
		
		JSeparator separator = new JSeparator();
		mnFile.add(separator);
		
		JMenuItem mntmOpenCms = new JMenuItem("Open CMS");
		mntmOpenCms.addMouseListener(new MouseAdapter() {
			@Override
			public void mousePressed(MouseEvent e) {
				OpenCMS();
			}
		});
		mnFile.add(mntmOpenCms);
		
		JMenuItem mntmSaveCms = new JMenuItem("Save CMS");
		mntmSaveCms.addMouseListener(new MouseAdapter() {
			@Override
			public void mousePressed(MouseEvent e) {
				SaveCMS();
			}
		});
		mnFile.add(mntmSaveCms);
		
		JSeparator separator_1 = new JSeparator();
		mnFile.add(separator_1);
		
		JMenuItem mntmExit = new JMenuItem("Exit");
		mntmExit.addMouseListener(new MouseAdapter() {
			@Override
			public void mousePressed(MouseEvent e) {
				dispose();
				System.exit(0);
			}
		});
		mnFile.add(mntmExit);
		
		JMenu mnHelp = new JMenu("Help");
		menuBar.add(mnHelp);
		
		JMenuItem mntmAbout = new JMenuItem("About");
		mntmAbout.addMouseListener(new MouseAdapter() {
			@Override
			public void mousePressed(MouseEvent e) {
				aboutClick();
			}
		});
		mnHelp.add(mntmAbout);
		
		JSplitPane splitPane = new JSplitPane();
		getContentPane().add(splitPane, BorderLayout.CENTER);
		splitPane.setDividerLocation(200);
		
		JScrollPane scrollPane = new JScrollPane();
		splitPane.setLeftComponent(scrollPane);
		
		rootNode = new DefaultMutableTreeNode("CMS");
		tree = new JTree(rootNode);
		tree.addTreeSelectionListener(new TreeSelectionListener() {
			public void valueChanged(TreeSelectionEvent e) {
				treeAfterSelect();
			}
		});
		tree.setRootVisible(false);
		scrollPane.setViewportView(tree);
		
		ctxMenu = new JPopupMenu();
		addPopup(tree, ctxMenu);
		
		JTabbedPane tabbedPane = new JTabbedPane(JTabbedPane.TOP);
		splitPane.setRightComponent(tabbedPane);
		
		JPanel panel = new JPanel();
		tabbedPane.addTab("Properties", null, panel, null);
		panel.setLayout(new BorderLayout(0, 0));
		
		JScrollPane scrollPane_1 = new JScrollPane();
		panel.add(scrollPane_1, BorderLayout.CENTER);
		
		lvProps = new JTable(new PropsListModel());
		lvProps.setFillsViewportHeight(true);
		scrollPane_1.setViewportView(lvProps);

		init();
	}

	protected void treeAfterSelect() {
		Object Tag = getSelectedNodeTag();
		if (Tag == null) return;
		if (Tag instanceof TElSignedCMSMessage)
		{
			DisplayCMSInfo((TElSignedCMSMessage)Tag);
			SetupPopupMenuCMS((TElSignedCMSMessage)Tag);
		}
		else if (Tag instanceof TElCMSContent)
		{
			SetupPopupMenuNothing();
		}
		else if (Tag instanceof TElCMSSignature)
		{
			DisplaySignatureInfo((TElCMSSignature)Tag);
			SetupPopupMenuSig((TElCMSSignature)Tag);
		}
		else if (Tag instanceof TElOCSPResponse)
		{
			DisplayOCSPResponseInfo((TElOCSPResponse)Tag);
			SetupPopupMenuOCSPResponse((TElOCSPResponse)Tag);
		}
		else if (Tag instanceof TElCMSTimestamp)
		{
			SetupPopupMenuTSP((TElCMSTimestamp)Tag);
		}
		else
			SetupPopupMenuNothing();
	}

	protected void aboutClick() {
		AboutDialog dlg = new AboutDialog();
		dlg.setModal(true);
		dlg.setVisible(true);
	}

	void init(){
		#error Please pick the evaluation license key from <SecureBlackbox>\LicenseKey.txt file and place it here. If the evaluation key expires, you can request an extension using the form on https://www.eldos.com/sbb/keyreq/
		SBUtils.setLicenseKey("<put the key here>");
		
        SBHTTPCRL.registerHTTPCRLRetrieverFactory();
        SBHTTPOCSPClient.registerHTTPOCSPClientFactory();

        FCMS = new TElSignedCMSMessage();
		FLog = "";
        FValidator = new TElX509CertificateValidator();        
	}

    private boolean ArrIsNullOrEmpty(byte[] B)
    {
        return ((B == null) || (B.length == 0));
    }

	private void NewCMS()
	{
		byte[] Data = SBUtils.bytesOfString("test data");
		FCMS.createNew(Data, 0, Data.length);
		RefreshCMS();
	}

	private void OpenCMS()
	{
		String fileName = getOpenFileName();
		if (fileName.length() > 0)
		{
			CloseActiveCMS();
			FCMSStream = new TElFileStream(fileName, "r", true);

			FCMS.open(FCMSStream, null, 0, 0);
			RefreshCMS();
		}
	}
	
	private void CloseActiveCMS()
	{
		if (FCMSStream != null)
		{
			FCMSStream.Free();
			FCMSStream =null;
		}
		if (FDataStream != null) 
		{
			FDataStream.Free();
			FDataStream = null;
		}
		tree.setRootVisible(false);
	}

	private void SaveCMS()
	{
		String fileName = getSaveFileName();
		if (fileName.length() > 0)
		{
			TElFileStream DestStream = new TElFileStream(fileName, "rw", true);
			try
			{
				FCMS.save(DestStream);
			}
			finally
			{
				DestStream.Free();
			}
		}
	}

	
	private String getSaveFileName() {
		JFileChooser fc = new JFileChooser();
	    
	    int returnVal = fc.showSaveDialog(this);
		if (returnVal == JFileChooser.APPROVE_OPTION)
	        return fc.getSelectedFile().getPath();

	    return "";
	}

	private void RefreshCMS() {
		rootNode.removeAllChildren();
		tree.setRootVisible(true);
		
		rootNode.setUserObject(new NodeData("CMS", FCMS));
		
		DefaultMutableTreeNode SubNode = new DefaultMutableTreeNode("Content");
		String s = "";
		if (FCMS.getContent() != null)
			s = FCMS.getContent().getSize() + " bytes";
		else
			s = "Not assigned";
		DefaultMutableTreeNode Node = new DefaultMutableTreeNode(s);
		Node.setUserObject(new NodeData(s, FCMS.getContent()));
		SubNode.setUserObject(new NodeData("Content", FCMS.getContent()));
		SubNode.add(Node);
		rootNode.add(SubNode);
		
		DefaultMutableTreeNode cSubNode = new DefaultMutableTreeNode("Certificates");
		for (int i = 0; i < FCMS.getCertificates().getCount(); i++)
		{
			s = FCMS.getCertificates().getCertificate(i).getSubjectName().CommonName;
			cSubNode.add(new DefaultMutableTreeNode("CN=" + s));
		}
		rootNode.add(cSubNode);

		DefaultMutableTreeNode sSubNode = new DefaultMutableTreeNode("Signatures");
		for (int i = 0; i < FCMS.getSignatureCount(); i++)
		{
			Node = new DefaultMutableTreeNode("Signature #" + i);
			RefreshSignature(FCMS.getSignature(i), Node, "Signature #" + i);
			sSubNode.add(Node);
		}
		rootNode.add(sSubNode);		
		
		((DefaultTreeModel)tree.getModel()).nodeStructureChanged(rootNode);
		
		expandTree();
	}
	
	private void expandTree() {	
	    int row = 0;
	    while (row < tree.getRowCount()) {
	      tree.expandRow(row);
	      row++;
	    }
	}
	
	private void RefreshSignature(TElCMSSignature Sig, DefaultMutableTreeNode V, String nodeTitle)
	{
		V.removeAllChildren();
		V.setUserObject(new NodeData(nodeTitle, Sig));
		DefaultMutableTreeNode Sub = new DefaultMutableTreeNode("Signer");
		V.add(Sub);
		DefaultMutableTreeNode SubSub = new DefaultMutableTreeNode("Issuer");
		Sub.add(SubSub);
		for (int i = 0; i < Sig.getSigner().getIssuer().getCount(); i++)
		{
			SubSub.add(new DefaultMutableTreeNode(SBStrUtils.oidToStr(Sig.getSigner().getIssuer().getOID(i)) + "=" +
				RDNValueToString(Sig.getSigner().getIssuer().getTag(i), Sig.getSigner().getIssuer().getValue(i))));
		}
		byte[] SN = Sig.getSigner().getSerialNumber();
		Sub.add(new DefaultMutableTreeNode("Serial=" + SBUtils.binaryToString(SN)));
		V.add(new DefaultMutableTreeNode("Digest algorithm=" + SBUtils.intToHex(Sig.getDigestAlgorithm(), 4)));
		if (Sig.getSigningCertificate().getIncluded())
		{
			Sub = new DefaultMutableTreeNode("Signing certificate chain info");
			V.add(Sub);
			for (int i = 0; i < Sig.getSigningCertificate().getCertIDCount(); i++)
			{
				DefaultMutableTreeNode Fin = new DefaultMutableTreeNode("Certificate #" + i);
				Sub.add(Fin);					
				RefreshCertRef(Sig.getSigningCertificate().getCertID(i), Fin, "Certificate #" + i);
			}
		}
		if (Sig.getCountersignatureCount() > 0)
		{
			Sub = new DefaultMutableTreeNode("Countersignatures");
			V.add(Sub);
			for (int i = 0; i < Sig.getCountersignatureCount(); i++)
			{
				DefaultMutableTreeNode Fin = new DefaultMutableTreeNode("Countersignature #" + i);
				Sub.add(Fin);					
				RefreshSignature(Sig.getCountersignature(i), Fin, "Countersignature #" + i);
			}
		}
		if (Sig.getTimestampCount() > 0)
		{
			Sub = new DefaultMutableTreeNode("Timestamps");
			V.add(Sub);
			for (int i = 0; i < Sig.getTimestampCount(); i++)
			{
				DefaultMutableTreeNode Fin = new DefaultMutableTreeNode("Timestamp #" + i);
				Sub.add(Fin);					
				RefreshTimestamp(Sig.getTimestamp(i), Fin);
			}
		}
		if (Sig.getCertificateRefs().getIncluded())
		{
			Sub = new DefaultMutableTreeNode("Certificate references");
			V.add(Sub);
			for (int i = 0; i < Sig.getCertificateRefs().getCertIDCount(); i++)
			{
				DefaultMutableTreeNode Fin = new DefaultMutableTreeNode("Ref #" + i);
				Sub.add(Fin);					
				RefreshCertRef(Sig.getCertificateRefs().getCertID(i), Fin, "Ref #" + i);
			}
		}
		if (Sig.getCertificateValues().getCount() > 0)
		{
			Sub = new DefaultMutableTreeNode("Certificates");
			V.add(Sub);
			for (int i = 0; i < Sig.getCertificateValues().getCount(); i++)
			{
				DefaultMutableTreeNode Fin = new DefaultMutableTreeNode("CN=" + Sig.getCertificateValues().getCertificate(i).getSubjectName().CommonName);
				Sub.add(Fin);					
			}
		}
		if (Sig.getRevocationRefs().getIncluded())
		{
			Sub = new DefaultMutableTreeNode("Revocation references");
			V.add(Sub);
			for (int i = 0; i < Sig.getRevocationRefs().getRefCount(); i++)
			{
				SubSub = new DefaultMutableTreeNode("Ref #" + i);
				Sub.add(SubSub);
				if (Sig.getRevocationRefs().getRevocationRef(i).getCRLIDCount() > 0)
				{
					DefaultMutableTreeNode Fin = new DefaultMutableTreeNode("CRL Identifiers");
					SubSub.add(Fin);
					for (int j = 0; j < Sig.getRevocationRefs().getRevocationRef(i).getCRLIDCount(); j++)
					{
						DefaultMutableTreeNode FinFin = new DefaultMutableTreeNode("CRL ID #" + j);
						Fin.add(FinFin);
						for (int k = 0; k < Sig.getRevocationRefs().getRevocationRef(i).getCRLID(j).getCRLIdentifier().getIssuer().getCount(); k++)
							FinFin.add(new DefaultMutableTreeNode(SBStrUtils.oidToStr(Sig.getRevocationRefs().getRevocationRef(i).getCRLID(j).getCRLIdentifier().getIssuer().getOID(k)) + "=" + 
								RDNValueToString(Sig.getRevocationRefs().getRevocationRef(i).getCRLID(j).getCRLIdentifier().getIssuer().getTag(k), Sig.getRevocationRefs().getRevocationRef(i).getCRLID(j).getCRLIdentifier().getIssuer().getValue(k))));
						FinFin.add(new DefaultMutableTreeNode("Time=" + Sig.getRevocationRefs().getRevocationRef(i).getCRLID(j).getCRLIdentifier().getTime().toString()));
						FinFin.add(new DefaultMutableTreeNode("Number=" + Sig.getRevocationRefs().getRevocationRef(i).getCRLID(j).getCRLIdentifier().getNumber()));
					}
				}
				if (Sig.getRevocationRefs().getRevocationRef(i).getOCSPIDCount() > 0)
				{
					DefaultMutableTreeNode Fin = new DefaultMutableTreeNode("OCSP Identifiers");
					SubSub.add(Fin);
					for (int j = 0; j < Sig.getRevocationRefs().getRevocationRef(i).getOCSPIDCount(); j++)
					{
						DefaultMutableTreeNode FinFin = new DefaultMutableTreeNode("OCSP ID #" + j);
						Fin.add(FinFin);
					}
				}
				if (!ArrIsNullOrEmpty(Sig.getRevocationRefs().getRevocationRef(i).getOtherRevType()))
				{
					SubSub.add(new DefaultMutableTreeNode(SBStrUtils.oidToStr(Sig.getRevocationRefs().getRevocationRef(i).getOtherRevType()) +
						"=" + SBUtils.binaryToString(Sig.getRevocationRefs().getRevocationRef(i).getOtherRevRefs())));
				}
			}
		}
		if (Sig.getRevocationValues().getIncluded())
		{
			Sub = new DefaultMutableTreeNode("Revocation values");
			V.add(Sub);
			if (Sig.getRevocationValues().getCRLCount() > 0)
			{
				for (int i = 0; i < Sig.getRevocationValues().getCRLCount(); i++)
				{
					SubSub = new DefaultMutableTreeNode("CRL #" + i);
					Sub.add(SubSub);
					RefreshCRL(Sig.getRevocationValues().getCRL(i), SubSub, "CRL #" + i);
				}
			}
			if (Sig.getRevocationValues().getOCSPResponseCount() > 0)
			{
				for (int i = 0; i < Sig.getRevocationValues().getOCSPResponseCount(); i++)
				{
					SubSub = new DefaultMutableTreeNode("OCSP #" + i);
					Sub.add(SubSub);
					RefreshOCSP(Sig.getRevocationValues().getOCSPResponse(i), SubSub, "OCSP #" + i);
				}
			}
			if (!ArrIsNullOrEmpty(Sig.getRevocationValues().getOtherRevType()))
			{
				Sub.add(new DefaultMutableTreeNode(SBStrUtils.oidToStr(Sig.getRevocationValues().getOtherRevType()) +
					"=" + SBUtils.binaryToString(Sig.getRevocationValues().getOtherRevValues())));
			}
		}
		if (Sig.getValidationTimestampCount() > 0)
		{
			Sub = new DefaultMutableTreeNode("Validation timestamps");
			V.add(Sub);
			for (int i = 0; i < Sig.getValidationTimestampCount(); i++)
			{
				SubSub = new DefaultMutableTreeNode("Timestamp #" + i);
				Sub.add(SubSub);
				RefreshTimestamp(Sig.getValidationTimestamp(i), SubSub);
			}
		}
	}

	private void AddNameValue(String Name, int Value)
	{
		AddNameValue(Name, Value + "");
	}
	
	private void AddNameValue(String Name, String Value)
	{
		PropsListModel model = (PropsListModel)lvProps.getModel();
		model.addItem(Name, Value, "");
	}

	void clearProps() {
		((PropsListModel)lvProps.getModel()).clear();
	}
	
	private void DisplayCMSInfo(TElSignedCMSMessage Msg)
	{
		clearProps();
		if (Msg.getContent() != null)
		{
			if (Msg.getContent().getSize() < 256)
			{
				byte[] buf = new byte[(int) Msg.getContent().getSize()];
				Msg.getContent().read(buf, 0, buf.length, 0);
				AddNameValue("Content", SBUtils.stringOfBytes(buf));
			}
			else
				AddNameValue("Content", "Content too long");
		}
		else
			AddNameValue("Content", "No content available");
		AddNameValue("Content-Type", SBStrUtils.oidToStr(Msg.getContentType()));
		AddNameValue("Signatures", Msg.getSignatureCount());
		AddNameValue("Certificates", Msg.getCertificates().getCount());
		AddNameValue("CRLs", Msg.getCRLs().getCount());   
	}

	private void DisplaySignatureInfo(TElCMSSignature Sig)
	{
		clearProps();
		AddNameValue("Digest algorithm", "0x" + SBUtils.intToHex(Sig.getDigestAlgorithm(), 4));
		AddNameValue("Public key algorithm", "0x" + SBUtils.intToHex(Sig.getPublicKeyAlgorithm(), 4));
		AddNameValue("Signature algorithm", "0x" + SBUtils.intToHex(Sig.getSignatureAlgorithm(), 4));
		if (!ArrIsNullOrEmpty(Sig.getContentType()))
			AddNameValue("Content-Type", SBStrUtils.oidToStr(Sig.getContentType()));
		if (new OleDate(Sig.getSigningTime()).toDouble() != 0)
			AddNameValue("Signing time", Sig.getSigningTime().toString());
		for (int i = 0; i < Sig.getCustomSignedAttributes().getCount(); i++)
			AddNameValue("Signed attribute #" + i, SBStrUtils.oidToStr(Sig.getCustomSignedAttributes().getAttribute(i)));
		for (int i = 0; i < Sig.getCustomUnsignedAttributes().getCount(); i++)
			AddNameValue("Unsigned attribute #" + i, SBStrUtils.oidToStr(Sig.getCustomUnsignedAttributes().getAttribute(i)));
		String s;
		if (Sig.getSignaturePolicy().getIncluded())
		{
			if (Sig.getSignaturePolicy().getImplied())
				s = "Implied";
			else
				s = SBStrUtils.oidToStr(Sig.getSignaturePolicy().getPolicyIdentifier());
			AddNameValue("Signature policy", s);
		}
		if (Sig.getContentHints().getIncluded())
		{
			AddNameValue("ContentHints.ContentType", SBStrUtils.oidToStr(Sig.getContentHints().getContentType()));
			AddNameValue("ContentHints.ContentDescription", Sig.getContentHints().getContentDescription());
		}
		if (!ArrIsNullOrEmpty(Sig.getContentIdentifier()))
			AddNameValue("Content Identifier", SBUtils.binaryToString(Sig.getContentIdentifier()));
		if (Sig.getContentReference().getIncluded())
		{
			AddNameValue("ContentReference.ContentType", SBStrUtils.oidToStr(Sig.getContentReference().getContentType()));
			AddNameValue("ContentReference.ContentIdentifier", SBUtils.binaryToString(Sig.getContentReference().getContentIdentifier()));
		}
		if (Sig.getCommitmentTypeIndication().getIncluded())
		{
			s = "Proof of: ";
			if (Sig.getCommitmentTypeIndication().getProofOfOrigin())
				s = s + "origin ";
			if (Sig.getCommitmentTypeIndication().getProofOfReceipt())
				s = s + "receipt ";
			if (Sig.getCommitmentTypeIndication().getProofOfDelivery())
				s = s + "delivery ";
			if (Sig.getCommitmentTypeIndication().getProofOfSender())
				s = s + "sender ";
			if (Sig.getCommitmentTypeIndication().getProofOfApproval())
				s = s + "approval ";
			if (Sig.getCommitmentTypeIndication().getProofOfCreation())
				s = s + "creation ";
			for (int i = 0; i < Sig.getCommitmentTypeIndication().getCustomIdentifiers().getCount(); i++)
				s = s + SBStrUtils.oidToStr((byte[])Sig.getCommitmentTypeIndication().getCustomIdentifiers().getItem(i)) + " ";
			AddNameValue("Commitment type", s);
		}
		if (Sig.getSignerLocation().getIncluded())
		{
			s = "";
			if (Sig.getSignerLocation().getCountryName().getValue().length() > 0)
				s = s + "C=" + Sig.getSignerLocation().getCountryName().getValue() + ", ";
			if (Sig.getSignerLocation().getLocalityName().getValue().length() > 0)
				s = s + "L=" + Sig.getSignerLocation().getLocalityName().getValue() + ", ";
			for (int i = 0; i < Sig.getSignerLocation().getPostalAddressLineCount(); i++)
				s = s + "P" + i + "=" + Sig.getSignerLocation().getPostalAddressLine(i).getValue() + ", ";
			if (s.length() > 0)
				s = s.substring(0, s.length() - 2);
			AddNameValue("Signer location", s);
		}
		if (Sig.getSigningCertificate().getIncluded())
			AddNameValue("Signing certificate", Sig.getSigningCertificate().getCertIDCount() + " identifiers");
		
		TSBCMSAdvancedSignatureValidity vres = Sig.getLastValidationResult();
		if (vres == TSBCMSAdvancedSignatureValidity.casvValid) 
		{
			s = "Valid";
		} else
		if (vres == TSBCMSAdvancedSignatureValidity.casvSignatureCorrupted) 
		{
			s = "Signature is corrupted";
		} else
		if (vres == TSBCMSAdvancedSignatureValidity.casvSignerNotFound) 
		{
			s = "Signing certificate was not found";
		} else
		if (vres == TSBCMSAdvancedSignatureValidity.casvIncompleteChain) 
		{
			s = "Incomplete certificate chain";
		} else
		if (vres == TSBCMSAdvancedSignatureValidity.casvBadCountersignature) 
		{
			s = "Bad countersignature(s)";
		} else
		if (vres == TSBCMSAdvancedSignatureValidity.casvBadTimestamp) 
		{
			s = "Bad timestamp(s)";				
		} else
		if (vres == TSBCMSAdvancedSignatureValidity.casvCertificateExpired) 
		{
			s = "One of certificates is expired or not yet valid";
		} else
		if (vres == TSBCMSAdvancedSignatureValidity.casvCertificateRevoked) 
		{
			s = "One of certificates is revoked";
		} else
		if (vres == TSBCMSAdvancedSignatureValidity.casvCertificateCorrupted) 
		{
			s = "One of certificates is corrupted";
		} else
		if (vres == TSBCMSAdvancedSignatureValidity.casvUntrustedCA) 
		{
			s = "One of CA certificates is not trusted";
		} else
		if (vres == TSBCMSAdvancedSignatureValidity.casvRevInfoNotFound) 
		{
			s = "Revocation information is incomplete";
		} else
		if (vres == TSBCMSAdvancedSignatureValidity.casvTimestampInfoNotFound) 
		{
			s = "Timestamp information is incomplete";
		} else
		if (vres == TSBCMSAdvancedSignatureValidity.casvFailure) 
		{
			s = "General failure";
		} else
		if (vres == TSBCMSAdvancedSignatureValidity.casvUnknown) 
		{
			s = "Unknown";
		} else
		{
			s = "Unknown";
		}
		AddNameValue("Last validation result", s);
	}

	private void DisplayOCSPResponseInfo(TElOCSPResponse Ocsp)
	{
		clearProps();
		AddNameValue("Signature algorithm", "0x" + SBUtils.intToHex(Ocsp.getSignatureAlgorithm(), 4));
		if (!ArrIsNullOrEmpty(Ocsp.getResponderID().getSHA1KeyHash()))
			AddNameValue("Responder SHA1 key hash", SBUtils.binaryToString(Ocsp.getResponderID().getSHA1KeyHash()));
		if (Ocsp.getResponderID().getName().getCount() > 0)
		{
			String s = "";
			for (int i = 0; i < Ocsp.getResponderID().getName().getCount(); i++)
				s = s + RDNValueToString(Ocsp.getResponderID().getName().getTag(i), Ocsp.getResponderID().getName().getValue(i)) + ", ";
			s = s.substring(0, s.length() - 2);
			AddNameValue("Responder name", s);
		}
		AddNameValue("Produced at", Ocsp.getProducedAt().toString());
		AddNameValue("Responses", Ocsp.getResponseCount());
	}	

	
	private String RDNValueToString(byte tag, byte[] value)
	{
		switch (tag)
		{
			case SBASN1Tree.SB_ASN1_NUMERICSTR:
			case SBASN1Tree.SB_ASN1_PRINTABLESTRING:
			case SBASN1Tree.SB_ASN1_IA5STRING:
			case SBASN1Tree.SB_ASN1_VISIBLESTRING:
			{
				return SBUtils.stringOfBytes(value);
			}
			case SBASN1Tree.SB_ASN1_UTF8STRING:
			{
				return new String(value);					
			}
			case SBASN1Tree.SB_ASN1_BMPSTRING:
			{
				if ((value.length >= 2) && (value[0] == 254) && (value[1] == 255))
					return new String(value, 2, value.length - 2);
				else
					if ((value.length >= 2) && (value[0] == 255) && (value[1] == 254))
					return new String(value, 2, value.length - 2);
				else
					return new String(value, 0, value.length);
			}
			case SBASN1Tree.SB_ASN1_UNIVERSALSTRING:
			{
				if ((value.length < 4) || (value[0] != 255) || (value[1] != 254) || (value[2] != 0) || (value[3] != 0))
					SBUtils.jSwapBigEndianDWords(value);

				String Result = new String(value, 0, value.length, Charset.forName("UTF-32"));
				if ((Result.length() > 0) && (Result.charAt(0) == (char)(0xFEFF)))
				{
					StringBuilder b = new StringBuilder(Result);
					b.delete(0, 1);
					Result = b.toString();
				}
				return Result;
			}
			default:
			{
				return SBUtils.stringOfBytes(value);
			}
		}
	}
	
	private void RefreshCertRef(TElCMSSignerIdentifier Ref, DefaultMutableTreeNode V, String nodeTitle)
	{
		V.removeAllChildren();
		V.setUserObject(new NodeData(nodeTitle , Ref));
		if (Ref.getIssuer().getCount() > 0)
		{
			DefaultMutableTreeNode Sub = new DefaultMutableTreeNode("Issuer");
			V.add(Sub);
			for (int k = 0; k < Ref.getIssuer().getName(0).getDirectoryName().getCount(); k++)
			{
				DefaultMutableTreeNode Node = new DefaultMutableTreeNode(SBStrUtils.oidToStr(Ref.getIssuer().getName(0).getDirectoryName().getOID(k)) + "=" +
					RDNValueToString(Ref.getIssuer().getName(0).getDirectoryName().getTag(k), Ref.getIssuer().getName(0).getDirectoryName().getValue(k)));
				Sub.add(Node);
			}
		}
		if (!ArrIsNullOrEmpty(Ref.getSerialNumber()))
		{
			DefaultMutableTreeNode Node = new DefaultMutableTreeNode("Serial number: " + SBUtils.binaryToString(Ref.getSerialNumber()));
			V.add(Node);
		}
		if (!ArrIsNullOrEmpty(Ref.getHash().getHashValue()))
		{
			DefaultMutableTreeNode Node = new DefaultMutableTreeNode("Certificate hash: " + SBUtils.binaryToString(Ref.getHash().getHashValue()));
			V.add(Node);
		}
	}

	private void RefreshTimestamp(TElCMSTimestamp Tsp, DefaultMutableTreeNode V)
	{
		V.removeAllChildren();
		V.setUserObject(Tsp);

		String s = "";
		
		if (Tsp.getTimestampType() == TSBCMSTimestampType.cvtGeneric) 
		{
			s = "Generic"; 
		}
		if (Tsp.getTimestampType() == TSBCMSTimestampType.cvtESC) 
		{ 
			s = "ES-C";
		}
		if (Tsp.getTimestampType() == TSBCMSTimestampType.cvtCertsAndCRLs) 
		{
			s = "Certs and CRLs"; 
		}
		if (Tsp.getTimestampType() == TSBCMSTimestampType.cvtArchive) 
		{
			s = "Archive"; 
		}

		DefaultMutableTreeNode Node = new DefaultMutableTreeNode("Type: " + s);
		V.add(Node);

		DefaultMutableTreeNode Sub = new DefaultMutableTreeNode("Certificates");
		V.add(Sub);
		for (int i = 0; i < Tsp.getCertificates().getCount(); i++)
		{
			Node = new DefaultMutableTreeNode(Tsp.getCertificates().getCertificate(i).getSubjectName().CommonName);
			Sub.add(Node);
		}

		Node = new DefaultMutableTreeNode("Timestamp: " + Tsp.getInfo().getTime().toString());
		V.add(Node);

		Sub = new DefaultMutableTreeNode("Signatures");
		V.add(Sub);
		for (int i = 0; i < Tsp.getSignatureCount(); i++)
		{
			DefaultMutableTreeNode SubSub = new DefaultMutableTreeNode("Signature #" + i);
			Sub.add(SubSub);
			RefreshSignature(Tsp.getSignature(i), SubSub, "Signature #" + i);
		}
	}	
	
	private void RefreshCRL(TElCertificateRevocationList Crl, DefaultMutableTreeNode V, String nodeTitle)
	{
		V.removeAllChildren();
		V.setUserObject(new NodeData(nodeTitle, Crl));
		DefaultMutableTreeNode Node;
		DefaultMutableTreeNode Sub = new DefaultMutableTreeNode("Issuer");
		V.add(Sub);
		for (int i = 0; i < Crl.getIssuer().getCount(); i++)
		{
			Node = new DefaultMutableTreeNode(SBStrUtils.oidToStr(Crl.getIssuer().getOID(i)) + "=" + RDNValueToString(Crl.getIssuer().getTag(i), Crl.getIssuer().getValue(i)));
			Sub.add(Node);
		}
		Node = new DefaultMutableTreeNode("This update: " + Crl.getThisUpdate());
		Sub.add(Node);
		Node = new DefaultMutableTreeNode("Next update: " + Crl.getNextUpdate());
		Sub.add(Node);
		Node = new DefaultMutableTreeNode("Signature algorithm: 0x: " + SBUtils.intToHex(Crl.getSignatureAlgorithm(), 4));
		Sub.add(Node);

		DefaultMutableTreeNode SubSub = new DefaultMutableTreeNode("Items");
		Sub.add(SubSub);
		for (int i = 0; i < Crl.getCount(); i++)
		{
			Node = new DefaultMutableTreeNode("Cert #" + i);
			SubSub.add(Node);
			Node.add(new DefaultMutableTreeNode(SBUtils.binaryToString(Crl.getItem(i).getSerialNumber())));
		}
	}

	private void RefreshOCSP(TElOCSPResponse Ocsp, DefaultMutableTreeNode V, String nodeTitle)
	{
		V.removeAllChildren();
		V.setUserObject(new NodeData(nodeTitle, Ocsp));
		DefaultMutableTreeNode Sub = new DefaultMutableTreeNode("Produced at " + Ocsp.getProducedAt());
		V.add(Sub);
		for (int i = 0; i < Ocsp.getResponseCount(); i++)
		{
			String SN = SBUtils.binaryToString(Ocsp.getResponse(i).getSerialNumber()) + ": ";
			switch (Ocsp.getResponse(i).getCertStatus())
			{
				case SBOCSPCommon.csGood: 
				{
					SN = SN + "Good";
					break;
				}
				case SBOCSPCommon.csRevoked: 
				{
					SN = SN + "Revoked";
					break;
				}
				case SBOCSPCommon.csUnknown: 
				{
					SN = SN + "Unknown";
					break;
				}
			}
			Sub.add(new DefaultMutableTreeNode(SN));
		}
		if (Ocsp.getCertificates().getCount() > 0)
		{
			Sub = new DefaultMutableTreeNode("Certificates");
			V.add(Sub);
			for (int i = 0; i < Ocsp.getCertificates().getCount(); i++)
			{
				Sub.add(new DefaultMutableTreeNode(Ocsp.getCertificates().getCertificate(i).getSubjectName().CommonName));
			}
		}
	}
	
	private static void addPopup(Component component, final JPopupMenu popup) {
		component.addMouseListener(new MouseAdapter() {
			public void mousePressed(MouseEvent e) {
				if (e.isPopupTrigger()) {
					showMenu(e);
				}
			}
			public void mouseReleased(MouseEvent e) {
				if (e.isPopupTrigger()) {
					showMenu(e);
				}
			}
			private void showMenu(MouseEvent e) {
				popup.show(e.getComponent(), e.getX(), e.getY());
			}
		});
	}
	
	private void SetupPopupMenuCMS(TElSignedCMSMessage Msg)
	{
		ctxMenu.removeAll();
		JMenuItem signItem = new JMenuItem("Sign");
		signItem.addMouseListener(new MouseAdapter() {
			@Override
			public void mousePressed(MouseEvent e) {
				HandleSignCMS(e);
			}
		});
		ctxMenu.add(signItem);
		JMenuItem setItem = new JMenuItem("Set content");
		setItem.addMouseListener(new MouseAdapter() {
			@Override
			public void mousePressed(MouseEvent e) {
				HandleSetContentCMS(e);
			}
		});
		ctxMenu.add(setItem);
	}

	NodeData getFirstNode(TreePath[] paths){
		Object obj = ((DefaultMutableTreeNode)paths[0].getLastPathComponent()).getUserObject();
		if (obj instanceof NodeData)
			return (NodeData)obj;
		return null;
	}
	
	protected String getOpenFileName() {
		JFileChooser fc = new JFileChooser();
	    
	    int returnVal = fc.showOpenDialog(this);
		if (returnVal == JFileChooser.APPROVE_OPTION)
	        return fc.getSelectedFile().getPath();

	    return "";
	}
	
	protected void HandleSetContentCMS(MouseEvent e) {
    	Object tag = getSelectedNodeTag();
		if ((tag != null) && tag instanceof TElSignedCMSMessage)
		{
			String fileName = getOpenFileName();
			if (fileName.length() > 0)
			{
				if (FDataStream != null)
				{
					FDataStream.Free();
					FDataStream = null;
				}
				FDataStream = new TElFileStream(fileName, "r", true);
				TElSignedCMSMessage CMS = (TElSignedCMSMessage)(tag);
				CMS.getContent().init(FDataStream, 0, FDataStream.getLength());
				RefreshCMS();
			}
		}
	}

	protected void HandleSignCMS(MouseEvent e) {
		Object tag = getSelectedNodeTag();
		if ((tag != null) && tag instanceof TElSignedCMSMessage)
		{		
			TElSignedCMSMessage CMS = (TElSignedCMSMessage)(tag);
			int Index = CMS.addSignature();
			RecalculateSignature(CMS.getSignature(Index));
			RefreshCMS();
		}
	}

	private boolean showConfirmMessage(String msg, String cap) {
		return (JOptionPane.showConfirmDialog(this, msg, cap, JOptionPane.YES_NO_OPTION) == JOptionPane.YES_OPTION);
	}
	
	private boolean RecalculateSignature(TElCMSSignature Sig)
	{
		boolean Result = false;
		if ((Sig.getTimestampCount() > 0) || (Sig.getCountersignatureCount() > 0) || (Sig.getValidationTimestampCount() > 0))
		{
			if (!(showConfirmMessage("All existing countersignatures and timestamps will be removed. Do you still wish to recalculate the signature?", "")))
				return false;
		}
		TElMemoryCertStorage Chain = null;
		TElX509Certificate Cert = RequestSigningCertificate(Chain);
		if (Cert != null)
		{
			try
			{
				// since recalculating the signature invalidates existing counter signatures
				// and time stamps, removing them.
				Sig.clearTimestamps();
				Sig.clearCountersignatures();
				Sig.clearValidationTimestamps();
				// setting up the necessary properties
				Sig.setSigningTime(new Date());
				Sig.getCommitmentTypeIndication().setIncluded(true);
				Sig.getCommitmentTypeIndication().setProofOfOrigin(true);
				Sig.getCommitmentTypeIndication().setProofOfCreation(true);
				Sig.setUsePSS(false);
				Sig.sign(Cert, Chain);
				Result = true;
			}
			finally
			{
				if (Chain != null)
					Chain.Free();
				Cert.Free();
			}
		}
		return Result;
	}
	
 	private String requestPassword() {
		JPasswordField jpf = new JPasswordField();
		int result = JOptionPane.showConfirmDialog(null, jpf, "Please enter password", JOptionPane.OK_CANCEL_OPTION);
		if (result == 0)
			return new String(jpf.getPassword());
		return "";
	}
	
	private TElX509Certificate RequestSigningCertificate(TElMemoryCertStorage Chain)
	{
		String fileName = getOpenFileName();
		if (fileName.length() > 0)
		{
			TElX509Certificate Result = new TElX509Certificate();
			try
			{
				TElFileStream F = new TElFileStream(fileName, "r", true);
				try
				{
					String Pass = requestPassword();
					int R = Result.loadFromStreamPFX(F, Pass, 0);
					if (R != 0) 
						throw new Exception("Failed to load PFX, error 0x" + SBUtils.intToHex(R, 4));
					if (showConfirmMessage("Do you wish to add the complete or partial certificate chain?", ""))
					{
						Chain = new TElMemoryCertStorage();
						TElX509Certificate Cert = LoadCert();
						while (Cert != null)
						{
							Chain.add(Cert, false);
							Cert = LoadCert();
						}
					}
					return Result;
				}
				finally
				{
					F.Free();
				}
			}
			catch (Exception Ex)
			{
				Result.Free();
				showErrorMessage(Ex);
			}
		}
		return null;
	}

	private void showErrorMessage(Exception ex) {
		JOptionPane.showMessageDialog(null, ex.getMessage(), "Error", JOptionPane.ERROR_MESSAGE);				
	}

	private void SetupPopupMenuSig(TElCMSSignature Sig)
	{
		ctxMenu.removeAll();
		JMenuItem vItem = new JMenuItem("Validate");
		vItem.addMouseListener(new MouseAdapter() {
			@Override
			public void mousePressed(MouseEvent e) {
				HandleValidateSig(e);
			}
		});
		ctxMenu.add(vItem);		

		JMenuItem rItem = new JMenuItem("Recalculate");
		rItem.addMouseListener(new MouseAdapter() {
			@Override
			public void mousePressed(MouseEvent e) {
				HandleRecalculateSig(e);
			}
		});		
		ctxMenu.add(rItem);
		
		JMenuItem tItem = new JMenuItem("Timestamp");
		tItem.addMouseListener(new MouseAdapter() {
			@Override
			public void mousePressed(MouseEvent e) {
				HandleTimestampSig(e);
			}
		});			
		ctxMenu.add(tItem);
		
		JMenuItem cItem = new JMenuItem("Countersign");
		cItem.addMouseListener(new MouseAdapter() {
			@Override
			public void mousePressed(MouseEvent e) {
				HandleCountersignSig(e);
			}
		});					
		ctxMenu.add(cItem);

		JMenuItem aItem = new JMenuItem("Add revocation information");
		ctxMenu.add(aItem);

		JMenuItem refSubItem = new JMenuItem("References only");
		refSubItem.addMouseListener(new MouseAdapter() {
			@Override
			public void mousePressed(MouseEvent e) {
				HandleAddRevRefsToSig(e);
			}
		});		
		aItem.add(refSubItem);
		
		JMenuItem refvSubItem = new JMenuItem("References and values");
		refvSubItem.addMouseListener(new MouseAdapter() {
			@Override
			public void mousePressed(MouseEvent e) {
				HandleAddRevValuesToSig(e);
			}
		});				
		aItem.add(refvSubItem);

		JMenuItem vtItem = new JMenuItem("Add validation timestamp");
		vtItem.addMouseListener(new MouseAdapter() {
			@Override
			public void mousePressed(MouseEvent e) {
				HandleAddValidationTimestampToSig(e);
			}
		});						
		ctxMenu.add(vtItem);
	}

	protected void HandleAddValidationTimestampToSig(MouseEvent e) {
    	Object tag = getSelectedNodeTag();
		if ((tag != null) && tag instanceof TElCMSSignature)
		{			
			TElCMSSignature Sig = (TElCMSSignature)(tag);
			String URL = requestURL("Please provide an URL of the TSP server", "URL request", "http://");
			try
			{
				TElHTTPTSPClient TSPCli = new TElHTTPTSPClient();
				TElHTTPSClient HTTPCli = new TElHTTPSClient();
				try
				{
					TSPCli.setHTTPClient(HTTPCli);
					TSPCli.setURL(URL);
					Sig.addValidationTimestamp(TSPCli, TSBCMSTimestampType.cvtArchive);
				}
				finally
				{
					HTTPCli.Destroy();
					TSPCli.Destroy();
				}
				RefreshCMS();
			}
			catch (Exception Ex)
			{
				showErrorMessage("Failed to timestamp the message: " + Ex.getMessage());
			}
			}
	}

	private String requestURL(String msg, String cap, String iURL) {
		JTextField jpf = new JTextField(iURL + msg);
		int result = JOptionPane.showConfirmDialog(null, jpf, cap, JOptionPane.OK_CANCEL_OPTION);
		if (result == 0)
			return jpf.getText();
		return "";
	}

	protected void HandleAddRevValuesToSig(MouseEvent e) {
    	Object tag = getSelectedNodeTag();
		if ((tag != null) && tag instanceof TElCMSSignature)
		{	
			TElCMSSignature Sig = (TElCMSSignature)(tag);
			AddRevInfo(Sig, false);				
		}
	}

	Object getSelectedNodeTag() {
	    TreePath[] paths = tree.getSelectionPaths();

	    if (paths != null && paths.length > 0)
	    {
	    	NodeData node = getFirstNode(paths);
	    	if (node == null) return null;
	    	return node.getTag();
	    }
	    
	    return null;	
	}
	
	private void AddRevInfo(TElCMSSignature Sig, Boolean RefsOnly)
	{
		RevInfoDialog frmRevInfoRequest = new RevInfoDialog();
		try
		{
			frmRevInfoRequest.setModal(true);
			frmRevInfoRequest.setVisible(true);
			if (frmRevInfoRequest.isOK())
			{
				TElMemoryCertStorage CertStorage = new TElMemoryCertStorage();
				TElMemoryCRLStorage CrlStorage = new TElMemoryCRLStorage();
				ArrayList OCSP = new ArrayList();
				try
				{
					for (int i = 0; i < frmRevInfoRequest.getItemsCount(); i++)
					{
						if (frmRevInfoRequest.getItemTag(i) instanceof TElX509Certificate)
							CertStorage.add((TElX509Certificate)(frmRevInfoRequest.getItemTag(i)), true);
						else if (frmRevInfoRequest.getItemTag(i) instanceof TElCertificateRevocationList)
							CrlStorage.add((TElCertificateRevocationList)(frmRevInfoRequest.getItemTag(i)));
						else if (frmRevInfoRequest.getItemTag(i) instanceof TElOCSPResponse)
							OCSP.add((TElOCSPResponse)frmRevInfoRequest.getItemTag(i));
					}
					Sig.addCompleteValidationData(CertStorage, CrlStorage, OCSP, RefsOnly);
					RefreshCMS();
				}
				finally
				{
					CertStorage.Free();
					CrlStorage.Free();
				}
			}
		}
		finally
		{
			frmRevInfoRequest.dispose();
		}
	}
	
	protected void HandleAddRevRefsToSig(MouseEvent e) {
		Object tag = getSelectedNodeTag();
		if ((tag != null) && tag instanceof TElCMSSignature)
		{
			TElCMSSignature Sig = (TElCMSSignature)(tag);
			AddRevInfo(Sig, true);
		}
	}

	protected void HandleCountersignSig(MouseEvent e) {
		Object tag = getSelectedNodeTag();
		if ((tag != null) && (tag instanceof TElCMSSignature))
		{
			TElCMSSignature Sig = (TElCMSSignature)tag;
			int Index = Sig.addCountersignature();
			try
			{
				if (!RecalculateSignature(Sig.getCountersignature(Index)))
					throw new Exception("Cancelled by user");
				RefreshCMS();
			}
			catch (Exception Ex)
			{
				Sig.removeCountersignature(Index);
				showErrorMessage("Countersigning failed: " + Ex.getMessage());
			}
		}
	}

	TSBCertificateValidateEvent.Callback onCertificateValidate = new TSBCertificateValidateEvent.Callback() {
		
		public void tsbCertificateValidateEventCallback(TObject sender,
				TElX509Certificate Certificate, TSBBoolean Validate) {
			
			TElHTTPSClient client = (TElHTTPSClient) sender;
			
			if ((Certificate.getChain() == null) || (Certificate.getChain().getCertificate(0) == Certificate))
			{				
				// For proper CRL and OCSP validation please read instructions in 
				// description of ElX509CertificateValidator class in the help file
				TElX509CertificateValidatorResult result = new TElX509CertificateValidatorResult();
				FValidator.validateForSSL(Certificate, client.getRemoteHost(), client.getRemoteIP(), TSBHostRole.hrServer, null, false, new Date(), result);
				Validate.value = (result.Validity == TSBCertificateValidity.cvOk) || (result.Validity == TSBCertificateValidity.cvSelfSigned);
			}
			else
				Validate.value = true; 
		}
	};

	protected void HandleTimestampSig(MouseEvent e) {
		Object tag = getSelectedNodeTag();
		if ((tag != null) && (tag instanceof TElCMSSignature))
		{		
			TElCMSSignature Sig = (TElCMSSignature)(tag);
			String URL = requestURL("Please provide an URL of the TSP server: ", "URL request", "http://");
			try
			{
				TElHTTPTSPClient TSPCli = new TElHTTPTSPClient();
				TElHTTPSClient HTTPCli = new TElHTTPSClient();
				try
				{
					HTTPCli.setOnCertificateValidate(new TSBCertificateValidateEvent(onCertificateValidate));		
					TSPCli.setHTTPClient(HTTPCli);
					TSPCli.setURL(URL);
					Sig.addTimestamp(TSPCli);
				}
				finally
				{
					HTTPCli.Destroy();
					TSPCli.Destroy();
				}
				RefreshCMS();
			}
			catch (Exception Ex)
			{
				showErrorMessage("Failed to timestamp the message: " + Ex.getMessage());
			}
		}		
	}

	protected void HandleRecalculateSig(MouseEvent e) {
		Object tag = getSelectedNodeTag();
		if ((tag != null) && (tag instanceof TElCMSSignature))
		{
			TElCMSSignature Sig = (TElCMSSignature)(tag);
			RecalculateSignature(Sig);
			RefreshCMS();
		}
	}

	protected void HandleValidateSig(MouseEvent e) {
		Object tag = getSelectedNodeTag();
		if ((tag != null) && (tag instanceof TElCMSSignature))
		{
			TElCMSSignature Sig = (TElCMSSignature)(tag);
			ValOptsDialog frmValidationOps = new ValOptsDialog();
			try
			{
				frmValidationOps.setModal(true);
				frmValidationOps.setVisible(true);
				if (frmValidationOps.isOK())
				{
					int Ops = 0;
					if (frmValidationOps.getRecursiveValidation())
						Ops = Ops | SBPKICommon.cvoRecursiveValidation;
					if (frmValidationOps.getValidateChains())
						Ops = Ops | SBPKICommon.cvoValidateChains;
					if (frmValidationOps.getValidateTimes())
						Ops = Ops | SBPKICommon.cvoValidateTimes;
					if (frmValidationOps.getCheckRevocationStatus())
						Ops = Ops | SBPKICommon.cvoCheckRevocationStatus;
					if (frmValidationOps.getRequireTimestamps())
						Ops = Ops | SBPKICommon.cvoRequireTimestamps;
					if (frmValidationOps.getIgnoreLocalTimestamps())
						Ops = Ops | SBPKICommon.cvoIgnoreLocalTimestamps;
					if (frmValidationOps.getValidateTrusts())
						Ops = Ops | SBPKICommon.cvoValidateTrusts;
					
					ClearLog();

					Sig.setOnIsCertificateTrusted(new TSBCMSCertificateTrustedEvent(HandleCertTrusted));
					Sig.setOnCertificateNeeded(new TSBCMSCertificateNeededEvent(HandleCertNeeded));
					Sig.setOnCRLNeeded(new TSBCMSCRLNeededEvent(HandleCRLNeeded));
					Sig.setOnOCSPResponseNeeded(new TSBCMSOCSPResponseNeededEvent(HandleOCSPNeeded));
					Sig.setOnItemValidated(new TSBCMSItemValidatedEvent(HandleItemValidated));

                    TSBCMSAdvancedSignatureValidity V = TSBCMSAdvancedSignatureValidity.casvUnknown;
                    if (frmValidationOps.getUseCertValidator())
                    {
                        FValidator.setCheckCRL(frmValidationOps.getCheckRevocationStatus());
                        FValidator.setCheckOCSP(frmValidationOps.getCheckRevocationStatus());
                        FValidator.setMandatoryCRLCheck(true);
                        FValidator.setMandatoryOCSPCheck(true);
                        V = Sig.validate(FValidator, (short)Ops);
                    }
                    else
                    {
                        V = Sig.validate((short)Ops);
                    }

					/*Sig.setOnIsCertificateTrusted(null);
					Sig.setOnCertificateNeeded(null);
					Sig.setOnCRLNeeded(null);
					Sig.setOnOCSPResponseNeeded(null);
					Sig.setOnItemValidated(null);*/

					String S = "";
					if (V == TSBCMSAdvancedSignatureValidity.casvValid) 
					{
						S = "Valid";
					}
					if (V == TSBCMSAdvancedSignatureValidity.casvSignatureCorrupted) 
					{
						S = "Signature is corrupted";					
					}
					if (V == TSBCMSAdvancedSignatureValidity.casvSignerNotFound) 
					{
						S = "Signing certificate was not found";
					}
					if (V == TSBCMSAdvancedSignatureValidity.casvIncompleteChain) 
					{
						S = "Incomplete certificate chain";
					}
					if (V == TSBCMSAdvancedSignatureValidity.casvBadCountersignature) 
					{
						S = "Bad countersignature(s)";
					}
					if (V == TSBCMSAdvancedSignatureValidity.casvBadTimestamp) 
					{
						S = "Bad timestamp(s)";
					}
					if (V == TSBCMSAdvancedSignatureValidity.casvCertificateExpired) 
					{
						S = "One of certificates is expired or not yet valid";
					}
					if (V == TSBCMSAdvancedSignatureValidity.casvCertificateRevoked) 
					{
						S = "One of certificates is revoked";
					}
					if (V == TSBCMSAdvancedSignatureValidity.casvCertificateCorrupted) 
					{
						S = "One of certificates is corrupted";
					}
					if (V == TSBCMSAdvancedSignatureValidity.casvUntrustedCA) 
					{
						S = "One of CA certificates is not trusted";
					}
					if (V == TSBCMSAdvancedSignatureValidity.casvRevInfoNotFound) 
					{
						S = "Revocation information is incomplete";
					}
					if (V == TSBCMSAdvancedSignatureValidity.casvTimestampInfoNotFound) 
					{
						S = "Timestamp information is incomplete";
					}
					if (V == TSBCMSAdvancedSignatureValidity.casvFailure) 
					{
						S = "General failure";
					}
					showMessage("Validity status: " + S);
					if (showConfirmMessage("Do you wish to view the verification log?", ""))
					{
						showMessage(FLog);
					}
				}
			}
			finally
			{
				frmValidationOps.dispose();
			}
		}		
	}

	TSBCMSCertificateTrustedEvent.Callback HandleCertTrusted = new TSBCMSCertificateTrustedEvent.Callback() {
		
		public void tsbcmsCertificateTrustedEventCallback(TObject sender,
				TElX509Certificate cert, TSBBoolean trusted) {
			if (cert.getSelfSigned())
				trusted.value = showConfirmMessage("Please confirm that certificate CN=" + cert.getSubjectName().CommonName + " is trusted.", "");
		}
	};
	
	TSBCMSCertificateNeededEvent.Callback HandleCertNeeded = new TSBCMSCertificateNeededEvent.Callback() {
		
		public TElX509Certificate tsbcmsCertificateNeededEventCallback(
				TObject sender, TElCertificateLookup lookup) {
			TElX509Certificate Cert = new TElX509Certificate();
			String CertInfo = "--------\r\n";
			if ((SBCustomCertStorage.lcIssuer & lookup.getCriteria()) > 0)
				CertInfo = CertInfo + "Issuer: " + RDNToString(lookup.getIssuerRDN()) + "\r\n";
			if ((SBCustomCertStorage.lcSubject & lookup.getCriteria()) > 0)
				CertInfo = CertInfo + "Subject: " + RDNToString(lookup.getSubjectRDN()) + "\r\n";
			if ((SBCustomCertStorage.lcValidity & lookup.getCriteria()) > 0)
				CertInfo = CertInfo + "Validity: " + lookup.getValidFrom() + ".." + lookup.getValidTo() + "\r\n";
			if ((SBCustomCertStorage.lcPublicKeyAlgorithm & lookup.getCriteria()) > 0)
				CertInfo = CertInfo + "Public key algorithm: " + lookup.getPublicKeyAlgorithm() + "\r\n";
			if ((SBCustomCertStorage.lcSignatureAlgorithm & lookup.getCriteria()) > 0)
				CertInfo = CertInfo + "Signature algorithm: " + lookup.getSignatureAlgorithm() + "\r\n";
			if ((SBCustomCertStorage.lcSerialNumber & lookup.getCriteria()) > 0)
				CertInfo = CertInfo + "Serial number: " + SBUtils.binaryToString(lookup.getSerialNumber()) + "\r\n";
			if ((SBCustomCertStorage.lcPublicKeyHash & lookup.getCriteria()) > 0)
				CertInfo = CertInfo + "Public key hash: " + SBUtils.binaryToString(lookup.getPublicKeyHash()) + "\r\n";
			CertInfo = CertInfo + "--------";
			if (showConfirmMessage("The following certificate is needed: \r\n\r\n" + CertInfo + "\r\n\r\n" + "Click \"Yes\" to load it or \"No\" if you do not have it.", ""))
			{
				Cert = LoadCert();
			}
			return Cert;
		}
	};
	
	private String RDNToString(TElRelativeDistinguishedName Rdn)
	{
		String Result = "";
		for (int i = 0; i < Rdn.getCount(); i++)
			Result = Result + SBStrUtils.oidToStr(Rdn.getOID(i)) + "=" + RDNValueToString(Rdn.getTag(i), Rdn.getValue(i)) + ", ";
		if (Rdn.getCount() > 0) 
			Result = Result.substring(0, Result.length() - 2);
		return Result;
	}
	
	TSBCMSCRLNeededEvent.Callback HandleCRLNeeded = new TSBCMSCRLNeededEvent.Callback() {
		
		public TElCertificateRevocationList tsbcmscrlNeededEventCallback(
				TObject sender, TElCMSCRLValidatedID ID, TElX509Certificate Cert) {
			TElCertificateRevocationList Crl = new TElCertificateRevocationList();
			if (showConfirmMessage("CRL containing revocation information for the certificate '" + Cert.getSubjectName().CommonName + "' is needed. Click 'Yes' to load it or 'No' if you do not have it.", ""))
			{
				String fileName = getOpenFileName();
				if (fileName.length() > 0)
				{
					try
					{
						FileInputStream F = new FileInputStream(fileName);
						try
						{
							assert(Crl.loadFromStream(F, 0) == 0);
						}
						finally
						{
							F.close();
						}
					}
					catch(Exception ex)
					{
						Crl.Free();
						Crl = null;
					}
				}
			}
			return Crl;
		}
	};
	
	TSBCMSOCSPResponseNeededEvent.Callback HandleOCSPNeeded = new TSBCMSOCSPResponseNeededEvent.Callback() {
		
		public TElOCSPResponse tsbcmsocspResponseNeededEventCallback(TObject sender,
				TElCMSOCSPResponsesID ID, TElX509Certificate Cert) {
			TElOCSPResponse Ocsp = new TElOCSPResponse();
			if (showConfirmMessage("OCSP response containing revocation information for the certificate '" + Cert.getSubjectName().CommonName + "' is needed. Click 'Yes' to load it or 'No' if you do not have it.", ""))
			{
				String fileName = getOpenFileName();
				if (fileName.length() > 0)
				{
					try
					{
						FileInputStream F = new FileInputStream(fileName);
						try
						{
							byte[] Buf = new byte[(int)new File(fileName).length()];
							int Size = F.read(Buf, 0, Buf.length);
							Ocsp.load(Buf, 0, Size);
						}
						finally
						{
							F.close();
						}
					}
					catch(Exception ex)
					{
						Ocsp = null;
					}
				}
			}
			return Ocsp;
		}
	};
	
	TSBCMSItemValidatedEvent.Callback HandleItemValidated = new TSBCMSItemValidatedEvent.Callback() {
		
		public void tsbcmsItemValidatedEventCallback(TObject sender,
				TSBCMSItemType ItemType, TObject Item, TSBBoolean Valid) {
			String S = "Validating ";
			if (ItemType == TSBCMSItemType.citCertificate) 
			{
				S = S + "certificate (" + ((TElX509Certificate)Item).getSubjectName().CommonName + ")";
			}
			if (ItemType == TSBCMSItemType.citCRL) 
			{
				S = S + "CRL (" + RDNToString(((TElCertificateRevocationList)Item).getIssuer()) + ")";
			}
			if (ItemType == TSBCMSItemType.citOCSPResponse) 
			{
				S = S + "OCSP response (" + RDNToString(((TElOCSPResponse)Item).getResponderID().getName()) + ")";
			}
			if (ItemType == TSBCMSItemType.citTimestamp) 
			{
				S = S + "Timestamp (" + ((TElCMSTimestamp)Item).getInfo().getTime().toString() + ")";
			}
			S = S + ", status=";
			if (Valid.value) S = S + "VALID"; 
			else S = S + "INVALID";
			AddToLog(S);
		}
	};
	
	private void AddToLog(String s)
	{
		FLog += s + "\r\n";
	}
	
	private void showMessage(String msg) {
		JOptionPane.showMessageDialog(null, msg, "Info", JOptionPane.INFORMATION_MESSAGE);		
	}

	private void ClearLog() {
		FLog = "";		
	}

	private void SetupPopupMenuTSP(TElCMSTimestamp Tsp)
	{
		ctxMenu.removeAll();
		JMenuItem Item = new JMenuItem("Validate");
		Item.addMouseListener(new MouseAdapter() {
			@Override
			public void mousePressed(MouseEvent e) {
				HandleValidateTSP(e);
			}
		});			
		ctxMenu.add(Item);
	}

	protected void HandleValidateTSP(MouseEvent e) {
		Object tag = getSelectedNodeTag();
		if ((tag != null) && (tag instanceof TElCMSTimestamp))
		{
			TElCMSTimestamp Tsp = (TElCMSTimestamp)(tag);
			if (Tsp.validate())
				showMessage("The timestamp corresponds to the timestamped document.");
			else
				showErrorMessage("The timestamp DOES NOT correspond to the timestamped document.");
		}		
	}

	private void SetupPopupMenuOCSPResponse(TElOCSPResponse Ocsp)
	{
		ctxMenu.removeAll();
		JMenuItem Item = new JMenuItem("Validate");
		Item.addMouseListener(new MouseAdapter() {
			@Override
			public void mousePressed(MouseEvent e) {
				HandleValidateOCSPResponse(e);
			}
		});				
		ctxMenu.add(Item);
	}

	protected void HandleValidateOCSPResponse(MouseEvent e) {
		Object tag = getSelectedNodeTag();
		if ((tag != null) && (tag instanceof TElOCSPResponse))
		{
			TElOCSPResponse Ocsp = (TElOCSPResponse)(tag);
			TSBCMSSignatureValidity V = Ocsp.validate();
			String S = "";
			if (V ==TSBCMSSignatureValidity.csvValid) 
			{
				S = "Valid";
			}
			if (V == TSBCMSSignatureValidity.csvInvalid) 
			{
				S = "Invalid";
			}
			if (V == TSBCMSSignatureValidity.csvSignerNotFound) 
			{
				S = "Signer not found";
			}
			if (V == TSBCMSSignatureValidity.csvGeneralFailure)
			{
				S = "General failure";
			}
			showMessage("Validity status: " + S);
		}
	}

	private void SetupPopupMenuNothing()
	{
		ctxMenu.removeAll();
	}

	private TElX509Certificate LoadCert()
	{
		String fileName = getOpenFileName();
		if (fileName.length() > 0)
		{
			try
			{
				FileInputStream F = new FileInputStream(fileName);
				try
				{
					TElX509Certificate TmpCert = new TElX509Certificate();
					try
					{
						switch (TElX509Certificate.detectCertFileFormat(F))
						{
							case SBX509.cfDER : 
							{
								TmpCert.loadFromStream(F, 0);
								break;
							}
							case SBX509.cfPEM : 
							{
								assert(TmpCert.loadFromStreamPEM(F, requestPassword(), 0) == 0);
								break;
							}
							case SBX509.cfPFX : 
							{
								assert(TmpCert.loadFromStreamPFX(F, requestPassword(), 0) == 0);
								break;
							}
							default:
								throw new Exception("Unsupported certificate format");
						}
						return TmpCert;
					}
					catch(Exception ex)
					{
						TmpCert.Free();
						throw ex;
					}
				}
				finally
				{
					F.close();
				}
			}
			catch(Exception ex)
			{
				showErrorMessage("Failed to load certificate");
			}
		}
		return null;
	}

	private void showErrorMessage(String msg) {
		JOptionPane.showMessageDialog(null, msg, "Error", JOptionPane.ERROR_MESSAGE);						
	}
	
}
