import java.net.InetAddress;
import java.util.Arrays;

import org.freepascal.rtl.TObject;

import SecureBlackbox.Base.*;
import SecureBlackbox.SSHCommon.*;

public class SshForwarding extends TObject
{
	public enum InState { Active, Closing, Closed };
	public enum OutState { Establishing, Active, Closing, Closed };
	
	TElSSHTunnelConnection	m_sshCon;
	TElSocket				m_socket;
	byte[]					m_socketToChannel = new byte[0];
	byte[]					m_channelToSocket = new byte[0];
	boolean					m_error;
	java.lang.Thread		m_forwardingLoop;
	private					InState m_inState;
	private					OutState m_outState;
	private String			m_tohost;
	private int				m_toport = 0;
	private int				m_received = 0;
	private int				m_sent = 0;
	Object					m_channelLock = new Object();
	Object					m_socketLock = new Object();
	ForwardEventListener m_listener;
	
	public SshForwarding(TElSSHTunnelConnection conn, String destHost, int destPort, ForwardEventListener l) 
	{
		m_listener = l;
		m_socket = null;
		m_sshCon = conn;
		SetupConnection();
		m_inState = InState.Active;
		m_outState = OutState.Establishing;
		m_tohost = destHost;
		m_toport = destPort;

		m_forwardingLoop = new java.lang.Thread(new Runnable() {
			public void run() {
				ForwardingThreadFunc();
			}
		});
        m_forwardingLoop.start();			
	}

	void ForwardingThreadFunc() 
	{
		byte[] buf = new byte[8192];
		TSBInteger len = new TSBInteger();
		TSBInteger index = new TSBInteger();
		TSBInteger sent = new TSBInteger();
		ArrayList socketList = new ArrayList();

		// connecting to remote side
		Log("Connecting to " + m_tohost + "...", false);
		m_socket = new TElSocket();
		try 
		{
			InetAddress addr = InetAddress.getByName(m_tohost);
			String addrStr = addr.getHostAddress();
			
			m_socket.setAddress(addrStr);
			m_socket.setPort(m_toport);

			m_socket.connect(10000);			
			m_outState = SshForwarding.OutState.Active;
		} 
		catch (Exception ex)
		{
            Log(ex.getMessage(), true);
			m_error = true;
		}
		m_listener.OnChange(this);
		
		while ((!m_error) && (!((m_inState == InState.Closed) && (m_outState == OutState.Closed)))) 
		{
			boolean changed = false;
			// socket operations
			if ((m_socket.getState() == SBSocket.issConnected) && (m_outState == OutState.Active))
			{
				// reading data from socket
				try 
				{
					socketList.clear();
					socketList.add(m_socket);
					len = new TSBInteger();
					//Socket.select(socketList, null, null, 1000);
					if (m_socket.canReceive(1000))
					{
						if (socketList.size() > 0) 
						{
							m_socket.receive(buf, buf.length, len);
							if (len.value > 0) 
							{
								WriteToChannelBuffer(buf, 0, len.value);
							} 
							else 
							{
								m_outState = OutState.Closed;
							}
							changed = true;
						} 
						else 
						{
	                        java.lang.Thread.sleep(0);
						}
					}
					// writing pending data to socket
					boolean received;
					do 
					{
						len.value = ReadFromSocketBuffer(buf, 0, buf.length);
						received = (len.value > 0);
						index = new TSBInteger();
						while (len.value > 0) 
						{
							m_socket.send(buf, index.value, len.value, sent);
							index.value += sent.value;
							len.value -= sent.value;
						}
						if (received) 
						{
                            changed = true;
						}
					} while (received);
				} 
				catch(Exception ex) 
				{
					Log(ex.getMessage(), true);
					m_outState = OutState.Closing;
					changed = true;
				}
			} 
			else if (!(m_socket.getState() == SBSocket.issConnected))
			{
				m_outState = OutState.Closed;
				changed = true;
			}
			// channel operations
			if (m_sshCon != null) 
			{
				do 
				{
					len.value = ReadFromChannelBuffer(buf, 0, buf.length);
					if (len.value > 0) 
					{
						m_listener.OnDataSend(this, buf, 0, len.value);
						m_sent += len.value;
						changed = true;
					}
				} while (len.value > 0);
			}						
			// re-adjusting states
			if ((m_inState == InState.Active) && ((m_outState == OutState.Closed) || (m_outState == OutState.Closing))) 
			{
				m_listener.OnClose(this);
				m_inState = InState.Closing;
				m_socket.Free();
				changed = true;
			} 
			else if (((m_outState == OutState.Active)) && ((m_inState == InState.Closing) || (m_inState == InState.Closed))) 
			{
				m_outState = OutState.Closing;
				changed = true;
			} 
			else if ((m_outState == OutState.Closing) && (!(m_socket.getState() == SBSocket.issConnected)))
			{
                m_outState = OutState.Closed;
				changed = true;
			}
			if (changed) 
			{
				m_listener.OnChange(this);
			}
			try {
				java.lang.Thread.sleep(0);
			} catch (InterruptedException e) {
				e.printStackTrace();
			}
		}
		if (m_socket.getState() == SBSocket.issConnected) 
		{
            m_socket.Free();
		}
		m_listener.OnFinish(this);
	}

	public void Close() 
	{
        m_inState = InState.Closed;
	}

	private void SetupConnection() 
	{
        m_sshCon.setOnClose(new TSSHChannelCloseEvent(m_sshConnection_OnClose));
		m_sshCon.setOnData(new TSSHDataEvent(m_sshConnection_OnData));
	}

	TSSHDataEvent.Callback m_sshConnection_OnData = new TSSHDataEvent.Callback() {
		
		public void tsshDataEventCallback(TObject sender, byte[] data) {
			m_received += data.length;
	        WriteToSocketBuffer(data, 0, data.length);
		}
	};
	
	TSSHChannelCloseEvent.Callback m_sshConnection_OnClose = new TSSHChannelCloseEvent.Callback() {
		
		public void tsshChannelCloseEventCallback(TObject sender, TSSHCloseType type) {
			m_inState = InState.Closed;
		}
	};

	public TElSSHTunnelConnection getSshConnection() {
		return m_sshCon;
	}
	
	public void setSshConnection(TElSSHTunnelConnection value) {
		m_sshCon = value; 
		SetupConnection();
		m_outState = OutState.Active;
		m_listener.OnChange(this);
	}
	

	public InState getInState() 
	{
		return m_inState;
	}

	public OutState getOutState() 
	{
		return m_outState;
	}

	public String getToHost()
	{
		return m_tohost;
	}

	public int getToPort()
	{
		return m_toport;
	}
	
	public int getReceived()
	{
		return m_received;
	}

	public int getSent()
	{
		return m_sent;
	}

	private void WriteToChannelBuffer(byte[] buffer, int offset, int length) 
	{
		synchronized(m_channelLock)
		{
			try 
			{
				byte[] newBuf = new byte[m_socketToChannel.length + length];
				newBuf = Arrays.copyOf(m_socketToChannel, m_socketToChannel.length);
				newBuf = Arrays.copyOf(buffer, length);
				m_socketToChannel = newBuf;
			} 
			finally 
			{
			}
		}
	}

	private void WriteToSocketBuffer(byte[] buffer, int offset, int length)
	{
		synchronized (m_socketLock){
			try 
			{
				byte[] newBuf = new byte[m_channelToSocket.length + length];
				newBuf = Arrays.copyOf(m_channelToSocket, m_channelToSocket.length);
				newBuf = Arrays.copyOf(buffer, length);
				m_channelToSocket = newBuf;
			} 
			finally 
			{
			}
		}
	}

	private int ReadFromChannelBuffer(byte[] buffer, int offset, int length)
	{
		int read = 0;
		synchronized (m_channelLock)
		{
			try 
			{
				read = Math.min(length, m_socketToChannel.length);
				buffer = Arrays.copyOf(m_socketToChannel, read);
				byte[] newBuf = new byte[m_socketToChannel.length - read];
				newBuf = Arrays.copyOf(m_socketToChannel, newBuf.length);
				m_socketToChannel = newBuf;
			} 
			finally 
			{
			}
		}
		return read;
	}

	private int ReadFromSocketBuffer(byte[] buffer, int offset, int length)
	{
		int read = 0;
		synchronized (m_socketLock)
		{
			try 
			{
				read = Math.min(length, m_channelToSocket.length);
				buffer = Arrays.copyOf(m_channelToSocket, read);
				byte[] newBuf = new byte[m_channelToSocket.length - read];
				newBuf = Arrays.copyOf(m_channelToSocket, newBuf.length);
				m_channelToSocket = newBuf;
			} 
			finally 
			{
			}
		}
		return read;
	}

	private void Log(String s, boolean error)
	{
		m_listener.OnLog(this, s, error);
	}

	public String getCurrentStatusState() {
		String result = "closed";
		
		if ((m_inState == InState.Active) && (m_outState == OutState.Active))
		{
			result = "active";
		} 
		else if (m_outState == OutState.Establishing) 
		{
			result = "establ";
		} 
		else if (m_inState == InState.Closing && m_outState == OutState.Closing)
		{
			result = "closing";
		}
        
        return result;
	}
}
