

import java.awt.BorderLayout;
import javax.swing.JButton;
import javax.swing.JDialog;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.border.EmptyBorder;
import javax.swing.border.TitledBorder;
import javax.swing.JTextField;
import javax.swing.JLabel;
import javax.swing.JTextArea;
import javax.swing.border.LineBorder;

import org.freepascal.rtl.TObject;

import SecureBlackbox.Base.*;
import SecureBlackbox.SSLClient.*;
import SecureBlackbox.SSLCommon.*;

import java.awt.Color;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import java.io.IOException;

public class Main extends JDialog {
	private static final long serialVersionUID = 1L;
	private final JPanel contentPanel = new JPanel();
	private TElSocket clientSocket;
	private TElSSLClient secureClient = null;	
	private TElMemoryCertStorage FCertStorage = null;
	private JTextField tbAddr;
	private JTextField tbPort;
	private JTextField tbLine;
	private JLabel lStatus;
	private JTextArea textArea;
	private JButton btnConnect;
	private JButton btnDisconnect;
	private JButton btnSend;
	private int FLastCert = 0;

	/**
	 * Launch the application.
	 */
	public static void main(String[] args) {
		try {
			Main dialog = new Main();
			dialog.setDefaultCloseOperation(JDialog.DISPOSE_ON_CLOSE);
			dialog.setVisible(true);
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	/**
	 * Create the dialog.
	 */
	public Main() {
		setTitle("SSL Chat Client");
		setBounds(100, 100, 470, 325);
		getContentPane().setLayout(new BorderLayout());
		contentPanel.setBorder(new EmptyBorder(5, 5, 5, 5));
		getContentPane().add(contentPanel, BorderLayout.CENTER);
		contentPanel.setLayout(null);
		
		JPanel panel = new JPanel();
		panel.setBorder(new TitledBorder(null, "Server address", TitledBorder.LEADING, TitledBorder.TOP, null, null));
		panel.setBounds(10, 11, 442, 52);
		contentPanel.add(panel);
		panel.setLayout(null);
		
		tbAddr = new JTextField();
		tbAddr.setText("127.0.0.1");
		tbAddr.setBounds(10, 21, 134, 20);
		panel.add(tbAddr);
		tbAddr.setColumns(10);
		
		tbPort = new JTextField();
		tbPort.setText("4567");
		tbPort.setBounds(154, 21, 53, 20);
		panel.add(tbPort);
		tbPort.setColumns(10);
		
		btnConnect = new JButton("Connect");
		btnConnect.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				connectClick();
			}
		});
		btnConnect.setBounds(223, 20, 91, 23);
		panel.add(btnConnect);
		
		btnDisconnect = new JButton("Disconnect");
		btnDisconnect.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				disconnectClick();
			}
		});
		btnDisconnect.setBounds(324, 20, 108, 23);
		panel.add(btnDisconnect);
		
		JLabel label = new JLabel(":");
		label.setBounds(147, 24, 22, 14);
		panel.add(label);
				
		tbLine = new JTextField();
		tbLine.setBounds(10, 238, 340, 20);
		contentPanel.add(tbLine);
		tbLine.setColumns(10);
		
		btnSend = new JButton("Send");
		btnSend.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				sendClick();
			}
		});
		btnSend.setBounds(361, 238, 91, 23);
		contentPanel.add(btnSend);
		
		lStatus = new JLabel("Started");
		lStatus.setBounds(10, 269, 442, 14);
		contentPanel.add(lStatus);
		
		textArea = new JTextArea();
		textArea.setBorder(new LineBorder(new Color(0, 0, 0)));
		textArea.setBounds(10, 66, 442, 161);
		contentPanel.add(textArea);
	
		init();
	}

	protected void connectClick() {
		reset();
		try
		{
			int port = Integer.parseInt(tbPort.getText());

			clientSocket = new TElSocket();
			
			clientSocket.setAddress(tbAddr.getText());
			clientSocket.setPort(port);
			
			clientSocket.connect(5000);

			if (clientSocket.getState() == SBSocket.issConnected) {
				secureClient.open();
				receive();
			} else {
				reset();
				setStatusText("Unable to connect to host");
			}
		} catch(Exception ex) {
			reset();
			setStatusText("Unable to connect to host");
			appendToMemo("Unable to connect to host - " + ex.getMessage());
		}		
	}

	private void receive() {
		java.lang.Thread t = new java.lang.Thread(new Runnable() {
			public void run() {
				processRequest();
			}
		});
		
		t.start();
	}

	private void processRequest() {	
		try
		{
			while (clientSocket.getState() == SBSocket.issConnected) 
			{
				if (clientSocket.canReceive(500) == true)
					secureClient.dataAvailable();
				else {
					java.lang.Thread.sleep(50);
				}
			}
			secureClient.close(false);
			clientSocket.close(false);
		} 
		catch(Exception e)
		{
			reset();
			setStatusText("Connection closed");
			if (e instanceof NullPointerException)
				appendToMemo("Connection closed");
			else
				appendToMemo("Connection closed - " + e.getMessage());
		}		
	}
	
	protected void sendClick() {
		secureClient.sendData(tbLine.getText().getBytes());
		StringBuilder sb = new StringBuilder(textArea.getText());
		sb.append("[CLIENT] ");
		sb.append(tbLine.getText());
		sb.append("\r\n");
		textArea.setText(sb.toString());
        tbLine.setText("");
    }

	protected void disconnectClick() {
		reset();
	}

	private void appendToMemo(String value) {
		StringBuilder sb = new StringBuilder(textArea.getText());
		sb.append(value);
		sb.append("\r\n");
		textArea.setText(sb.toString());
		
	}

	private void setStatusText(String value) {
		lStatus.setText(value);
	}

	private void reset() {
		if (secureClient.getActive())
			secureClient.close(false);

		if (clientSocket != null)
		{
			clientSocket.close(false);
			clientSocket = null;
		}
		
		if (FCertStorage != null)
		  FCertStorage.clear();
		FCertStorage = null;

		btnConnect.setEnabled(true);
		btnDisconnect.setEnabled(false);
		tbAddr.setEnabled(true);
		tbPort.setEnabled(true);	
		btnSend.setEnabled(true);
		setStatusText("");		
	}
	
	private void init()
	{
		#error Please pick the evaluation license key from <SecureBlackbox>\LicenseKey.txt file and place it here. If the evaluation key expires, you can request an extension using the form on https://www.eldos.com/sbb/keyreq/
		SBUtils.setLicenseKey("<put the key here>");

		secureClient = new TElSSLClient(null);	
		secureClient.setEnabled(true);
		secureClient.setVersions((short) (SBSSLConstants.sbTLS12 |SBSSLConstants.sbTLS11 | SBSSLConstants.sbTLS1));
		
		secureClient.setOnOpenConnection(new TSBOpenConnectionEvent(ElSecureClientOpenConnection));
        secureClient.setOnCloseConnection(new SecureBlackbox.SSLCommon.TSBCloseConnectionEvent(ElSecureClientCloseConnection));
        secureClient.setOnData(new TSBDataEvent(ElSecureClientOnData));
        secureClient.setOnSend(new TSBSendEvent(ElSecureClientSend));
        secureClient.setOnReceive(new TSBReceiveEvent(ElSecureClientReceive));
        secureClient.setOnCertificateValidate(new TSBCertificateValidateEvent(ElSecureClientCertificateValidate));
        secureClient.setOnCertificateNeededEx(new TSBCertificateNeededExEvent(ElSecureClientCertificateNeededEx));
		secureClient.setOnError(new TSBErrorEvent(OnError));
	}
	
	protected void showErrorMessage(Exception e) {
		JOptionPane.showMessageDialog(null, e.getMessage(), "Error", JOptionPane.ERROR_MESSAGE);
	}
	
	private String getCurrentPath() {
		try {
			return new java.io.File(".").getCanonicalPath();
		} catch (IOException e) {
			showErrorMessage(e);
		}
		return ".";
	}
	
	// event's
	
	TSBOpenConnectionEvent.Callback ElSecureClientOpenConnection = new TSBOpenConnectionEvent.Callback() {
		public void tsbOpenConnectionEventCallback(TObject sender) {		
			btnSend.setEnabled(true);
			btnConnect.setEnabled(false);
			btnDisconnect.setEnabled(true);
			tbAddr.setEnabled(false);
			tbPort.setEnabled(false);	

			StringBuilder sb = new StringBuilder(textArea.getText());
			sb.append("Secure Connection Established. SSL version is ");
			if (secureClient.getCurrentVersion() == SBSSLConstants.sbSSL2)
				sb.append("SSL2");
			else if (secureClient.getCurrentVersion() == SBSSLConstants.sbSSL3)
				sb.append("SSL3");
			else if (secureClient.getCurrentVersion() == SBSSLConstants.sbTLS1)
				sb.append("TLS1");
			else if (secureClient.getCurrentVersion() == SBSSLConstants.sbTLS11)
				sb.append("TLS1.1");
			sb.append("\r\n");
			textArea.setText(sb.toString());
			setStatusText("Secure Connection Established");		
		}
	};
	
	SecureBlackbox.SSLCommon.TSBCloseConnectionEvent.Callback ElSecureClientCloseConnection = new SecureBlackbox.SSLCommon.TSBCloseConnectionEvent.Callback() {
		public void tsbCloseConnectionEventCallback(TObject sender, SecureBlackbox.SSLCommon.TSBCloseReason reason) {		
			reset();
			setStatusText("Secure Connection Closed");
		}
	};
	
	TSBDataEvent.Callback ElSecureClientOnData = new TSBDataEvent.Callback() {
		public void tsbDataEventCallback(TObject sender, byte[] data) {		
			StringBuilder sb = new StringBuilder(textArea.getText());
			sb.append("[SERVER] ");
			sb.append(new String(data));
			sb.append("\r\n");
			textArea.setText(sb.toString());		
		}
	};
	
	TSBSendEvent.Callback ElSecureClientSend = new TSBSendEvent.Callback() {
		public void tsbSendEventCallback(TObject sender, byte[] buffer) {		
			try
			{	
				if (clientSocket != null) {
					if (buffer.length > 0) {
						TSBInteger sent = new TSBInteger();
						clientSocket.send(buffer, 0, buffer.length, sent);
					}
				}
			}
			catch (Exception ex)
			{
				reset();
				setStatusText("Connection closed");
				appendToMemo("Connection closed - " + ex.getMessage());
			}		
		}
	};
	
	TSBReceiveEvent.Callback ElSecureClientReceive = new TSBReceiveEvent.Callback() {		
		public void tsbReceiveEventCallback(TObject sender, byte[] buffer, int maxSize, TSBInteger written) {
			try
			{
				int err = clientSocket.receive(buffer, maxSize, written);
				if (err != 0)
					written.value = 0;
			}
			catch (Exception ex)
			{
				written.value = 0;
			}
		}
	};
	
	TSBCertificateValidateEvent.Callback ElSecureClientCertificateValidate = new TSBCertificateValidateEvent.Callback() {
		public void tsbCertificateValidateEventCallback(TObject sender, TElX509Certificate cert, TSBBoolean accept) {
			// NEVER do this in real life since this makes security void. 
			// Instead validate the certificate as described on http://www.eldos.com/sbb/articles/1966.php
			accept.value = true;
		}
	};
	
	TSBCertificateNeededExEvent.Callback ElSecureClientCertificateNeededEx = new TSBCertificateNeededExEvent.Callback() {
		public TElX509Certificate tsbCertificateNeededExEventCallback(TObject sender) {
			TElX509Certificate Certificate;
			if (FCertStorage == null)
			{
				FCertStorage = new TElMemoryCertStorage();
				SelectCertDialog frm = new SelectCertDialog();
				try
				{
					frm.setMode(TSelectCertMode.ClientCert);
					try {
						frm.loadStorage(getCurrentPath() + "/src/CertStorageDef.ucs", FCertStorage);
					} catch (Exception e) {
						showErrorMessage(e);
					}
					frm.setStorage(FCertStorage);
					frm.setModal(true);
					frm.setVisible(true);
					if (frm.isOk())
					{
						FCertStorage = frm.getStorage(FCertStorage);
					}
					else
						FCertStorage.clear();
				}
				finally
				{
					frm.dispose();
				}

				FLastCert = -1;
			}

			FLastCert++;
			if (FLastCert >= FCertStorage.getCount())
			{
				Certificate = null;
			}
			else
				Certificate = FCertStorage.getCertificate(FLastCert);
			
			return Certificate;
		}
	};
	
	TSBErrorEvent.Callback OnError = new TSBErrorEvent.Callback() {
		public void tsbErrorEventCallback(TObject sender, int errorCode, boolean fatal, boolean remote) {
			StringBuilder sb = new StringBuilder(textArea.getText());
			sb.append("SSL protocol error " + errorCode);
			textArea.setText(sb.toString());			
		}
	};
}
