import java.awt.BorderLayout;
import java.awt.FlowLayout;

import javax.swing.JButton;
import javax.swing.JDialog;
import javax.swing.JPanel;
import javax.swing.ListSelectionModel;
import javax.swing.border.EmptyBorder;
import javax.swing.JTable;
import javax.swing.JScrollPane;

import SecureBlackbox.Base.SBConstants;
import SecureBlackbox.Base.SBUtils;
import SecureBlackbox.Base.TElClientTSPInfo;
import SecureBlackbox.Base.TElX509Certificate;
import SecureBlackbox.Base.TElX509CertificateValidator;
import SecureBlackbox.Base.TElX509CertificateValidatorResult;
import SecureBlackbox.Base.TSBCertificateValidity;
import SecureBlackbox.XML.TElXMLDOMElement;
import SecureBlackbox.XMLSecurity.SBXMLAdESIntf;
import SecureBlackbox.XMLSecurity.SBXMLSec;
import SecureBlackbox.XMLSecurity.TElXAdESVerifier;
import SecureBlackbox.XMLSecurity.TElXMLCertID;
import SecureBlackbox.XMLSecurity.TSBXAdESValidity;

import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import java.util.Arrays;
import java.util.Date;

public class XADESInfoFrm extends JDialog {

	private final JPanel contentPanel = new JPanel();
	private JTable table;
	
	private TElX509CertificateValidator certValidator = new TElX509CertificateValidator();

	/**
	 * Create the dialog.
	 */
	public XADESInfoFrm() {
		setTitle("Signature Handler Info");
		setBounds(100, 100, 505, 388);
		getContentPane().setLayout(new BorderLayout());
		contentPanel.setBorder(new EmptyBorder(5, 5, 5, 5));
		getContentPane().add(contentPanel, BorderLayout.CENTER);
		contentPanel.setLayout(null);
		
		JScrollPane scrollPane = new JScrollPane();
		scrollPane.setBounds(10, 10, 469, 296);
		contentPanel.add(scrollPane);
		
		table = new JTable(new XADESInfoModel());
		table.setFillsViewportHeight(true);
		table.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
		scrollPane.setViewportView(table);
		
		{
			JPanel buttonPane = new JPanel();
			buttonPane.setLayout(new FlowLayout(FlowLayout.RIGHT));
			getContentPane().add(buttonPane, BorderLayout.SOUTH);
			{
				JButton cancelButton = new JButton("Cancel");
				cancelButton.addActionListener(new ActionListener() {
					public void actionPerformed(ActionEvent e) {
						setVisible(false);
					}
				});
				cancelButton.setActionCommand("Cancel");
				buttonPane.add(cancelButton);
			}
		}
	}

	public void init(TElXAdESVerifier verifier, TSBXAdESValidity validity, int reasons, Date validationMoment) 
	{
		XADESInfoModel model = (XADESInfoModel) table.getModel();
		model.clear();
		
		if (verifier == null || !verifier.getIsEnabled())
			return;
		
		if (validity == TSBXAdESValidity.xsvValid)
			model.add("", "Validation results", "XAdES info validated successfully");
		else if (validity == TSBXAdESValidity.xsvIncomplete)
			model.add("", "Validation results", "Incomplete validation of XAdES info");
		else
			model.add("", "Validation results", "Failed to validate XAdES info");
		
		if (reasons != 0)
			model.add("", "Reasons", Utils.xadesReasonsToString(reasons));
		
		model.add("", "", "");
		model.add("", "XAdES Version", Utils.xadesVersionToString(verifier.getXAdESVersion()));
		model.add("", "XAdES Form", Utils.xadesFormToString(verifier.getXAdESForm()));
		
		if ((verifier.getIncluded() | SBXMLAdESIntf.xipProductionPlace) == SBXMLAdESIntf.xipProductionPlace) {
			if (verifier.getProductionPlace().getCountryName().length() > 0)
				model.add("XAdES_BES", "ProductionPlace.CountryName", verifier.getProductionPlace().getCountryName());
			if (verifier.getProductionPlace().getStateOrProvince().length() > 0)
				model.add("XAdES_BES", "ProductionPlace.StateOrProvince", verifier.getProductionPlace().getStateOrProvince());
			if (verifier.getProductionPlace().getPostalCode().length() > 0)
				model.add("XAdES_BES", "ProductionPlace.PostalCode", verifier.getProductionPlace().getPostalCode());
			if (verifier.getProductionPlace().getCity().length() > 0)
				model.add("XAdES_BES", "ProductionPlace.City", verifier.getProductionPlace().getCity());
		}
		
		if ((verifier.getIncluded() | SBXMLAdESIntf.xipSignerRole) == SBXMLAdESIntf.xipSignerRole) {
			for (int i = 0; i < verifier.getSignerRole().getClaimedRoles().getCount(); i++) {
				if (verifier.getSignerRole().getClaimedRoles().getItem(i).getValue() instanceof TElXMLDOMElement)
					model.add("XAdES_BES", "SignerRole.ClaimedRoles[]", ((TElXMLDOMElement)verifier.getSignerRole().getClaimedRoles().getItem(i).getValue()).getInnerXML());
			}
		}
		
		if (verifier.getSigningTime() != null)
			model.add("XAdES_BES", "Signing Time", SBUtils.dateTimeToUTCTime(verifier.getSigningTime(), true) + " UTC");
		else
			model.add("XAdES_BES", "Signing Time", "none");
		
		if (verifier.getCertIDCount() > 0) {
			for (int i = 0; i < verifier.getCertIDCount(); i++) {
				TElXMLCertID certId = verifier.getCertIDs(i);
				String s = Utils.getOIDValue(certId.getIssuerSerial().getIssuerRDN(), SBConstants.SB_CERT_OID_COMMON_NAME.Data, "/");
				
				if (s.length() == 0)
					s = Utils.getOIDValue(certId.getIssuerSerial().getIssuerRDN(), SBConstants.SB_CERT_OID_ORGANIZATION.Data, "/");
				
				if (s.length() == 0)
					s = "<unknown>";
				
				byte[] b = certId.getIssuerSerial().getSerialNumber();
				model.add("XAdES_BES", "SigningCertificates[].SerialNumber", SBUtils.beautifyBinaryString(SBUtils.binaryToString(b), ':'));
				model.add("XAdES_BES", "SigningCertificates[].Issuer", s);
			}
		}
		
		for (int i = 0; i < verifier.getAllDataObjectsTimestampCount(); i++)
			addTimestampDetails(verifier, verifier.getAllDataObjectsTimestamps(i), validationMoment, "XADES_BES", "All Data Objects Timestamp");
		
		if (verifier.getPolicyId() != null) {
			if (verifier.getPolicyId().getSigPolicyId().getDescription().length() > 0)
				model.add("XAdES_EPES", "SigPolicyId.Description", verifier.getPolicyId().getSigPolicyId().getDescription());
			if (verifier.getPolicyId().getSigPolicyId().getDocumentationReferences().getText().length() > 0)
				model.add("XAdES_EPES", "SigPolicyId.Description", verifier.getPolicyId().getSigPolicyId().getDocumentationReferences().getText());
			if (verifier.getPolicyId().getSigPolicyId().getIdentifier().length() > 0) {
				model.add("XAdES_EPES", "SigPolicyId.Identifier", verifier.getPolicyId().getSigPolicyId().getIdentifier());
				try {
					byte[] buf = Utils.downloadData(verifier.getPolicyId().getSigPolicyId().getIdentifier());
					buf = SBXMLSec.calculateDigest(buf, SBXMLSec.uriToDigestMethod(verifier.getPolicyId().getSigPolicyHash().getDigestMethod()));
					
					if (Arrays.equals(buf, verifier.getPolicyId().getSigPolicyHash().getDigestValue()))
						model.add("XAdES_EPES", "Signature Policy", "Policy data is valid");
					else
						model.add("XAdES_EPES", "Signature Policy", "INVALID POLICY DATA");
				} catch (Exception ex) {
					ex.printStackTrace();
					model.add("XAdES_EPES", "Signature Policy", "Failed to load policy data");
				}
			}
		}
		
		if (verifier.getSignatureTimestampCount() > 0 || verifier.getSigAndRefsTimestampCount() > 0 || verifier.getArchiveTimestampCount() > 0) {
			if (verifier.getCertifiedSigningTime() != null)
				model.add("XAdES_T", "Certified Signing Time", SBUtils.dateTimeToUTCTime(verifier.getCertifiedSigningTime(), true) + " UTC");
			else
				model.add("XAdES_T", "Certified Signing Time", "none");
		}
		
		for (int i = 0; i < verifier.getSignatureTimestampCount(); i++)
			addTimestampDetails(verifier, verifier.getSignatureTimestamps(i), validationMoment, "XAdES_T", "Signature Timestamp");
		for (int i = 0; i < verifier.getRefsOnlyTimestampCount(); i++)
			addTimestampDetails(verifier, verifier.getRefsOnlyTimestamps(i), validationMoment, "XAdES_X", "References Only Timestamp");
		for (int i = 0; i < verifier.getSigAndRefsTimestampCount(); i++)
			addTimestampDetails(verifier, verifier.getSigAndRefsTimestamps(i), validationMoment, "XAdES_X", "Signature And References Timestamp");
		for (int i = 0; i < verifier.getArchiveTimestampCount(); i++)
			addTimestampDetails(verifier, verifier.getArchiveTimestamps(i), validationMoment, "XAdES_A", "Archive Timestamp");
		
		if (verifier.getArchiveTimestampCount() > 0) {
			if (verifier.getLastArchivalTime() != null)
				model.add("XAdES_A", "Last Archival Time", SBUtils.dateTimeToUTCTime(verifier.getLastArchivalTime(), true) + " UTC");
			else
				model.add("XAdES_A", "Last Archival Time", "none");
		}
	}
	
	private void addTimestampDetails(TElXAdESVerifier verifier, TElClientTSPInfo tspInfo, Date validationMoment, String form, String name) {
		XADESInfoModel model = (XADESInfoModel) table.getModel();
		
		if (verifier.checkTimestamp(tspInfo)) {
			TElX509Certificate cert = tspInfo.getSignerCertificate();
			TSBCertificateValidity validity = TSBCertificateValidity.cvOk;
			int reason = 0;
			
			if (cert != null) {
				TElX509CertificateValidatorResult res = new TElX509CertificateValidatorResult();
				certValidator.validate(cert, tspInfo.getCertificates(), false, validationMoment, res);
				validity = res.Validity;
				reason = res.Reason;
			}
			
			if (validity == TSBCertificateValidity.cvOk) {
				model.add(form, name, SBUtils.dateTimeToUTCTime(tspInfo.getTime(), true) + " UTC");
				model.add(form, name, SBUtils.beautifyBinaryString(SBUtils.binaryToString(tspInfo.getSerialNumber()), ':'));
			} else {
				model.add(form, name, String.format("%s (Reason: %s)", Utils.x509ValidityToString(validity), Utils.x509ReasonToString(reason)));
			}
		} else {
			model.add(form, name, "INVALID TIMESTAMP");
		}
	}
}
